-- LuaOrbCfg.verbose_mode = 1

-- Auxiliary function

function table2str(table)
   local i,v = next(table,nil)
   local result = "{\n"
   while i do
      if type(i) == "string" then
         result = result .. "   " .. i .. " = "
      else
         result = result .. "   [" .. i .. "] = "
      end
      if type(v) == "table" then
         result = result .. table2str(v)
      elseif type(v) == "number" then
         result = result .. v
      else
         result = result .. "[[" .. v .."]]"
      end
      i,v = next(table,i)
      if i then
         result = result .. ",\n"
      else
         result = result .. "}"
      end
   end
   return result
end



function tstException_exception_handler(self, op, ex_name, ex_descr)
   return "exception"
end


function do_test(file)
   if type(file) == "string" then writeto(file) end

   write("---------------------\n")
   local tstAttrib = tst:getTstAttrib()
   assert(tstAttrib.domain == "inf.puc-rio.br")
   tstAttrib.host = "obaluae"
   assert(tstAttrib.host == "obaluae")
   write ("Attributes OK\n")

   tst:print("---------------------")
   tst:print(tstAttrib.host.."."..tstAttrib.domain)

   write("---------------------\n")
   local tstStructSeq = tst:getTstStructSeq()
   assert (tstStructSeq:add_book({title="T1", author="A1"}))
   assert (not tstStructSeq:add_book({title="T1", author="A1"})) 


   local booklist = {
      {title="T2", author="A2"},
      {title="T3", author="A3"},
      {title="T4", author="A4"},
      {title="T5", author="A1"},
      {title="T6", author="A1"},
      {title="T7", author="A2"},
      {title="T8", author="A3"},
   }

   assert (tstStructSeq:add_books(booklist)) 


   tinsert(booklist,1,{title="T1", author="A1"})

   assert (isSimilar(tstStructSeq:get_allbooks(),booklist))

   assert (isEqual(tstStructSeq:get_book("T3"),{title="T3", author="A3"}))

   local flag, list = tstStructSeq:get_booklist("A1")

   assert (isSimilar(list,{{title="T1", author="A1"},{title="T5", author="A1"},{title="T6", author="A1"}}))

   local flag, list = tstStructSeq:get_booklist("A10")
   assert (not flag)

   write("Structs OK\n")


   write("---------------------\n")
   local tstStructArray = tst:getTstStructArray()

   local p1 = {x=10,y=20}
   local p2 = {x=30,y=40}


   assert (isEqual (tstStructArray:draw_pt(p1), p1))
   assert (isEqual (tstStructArray:draw_pt(p2), p2))

   local rec = tstStructArray:create_rect(p1,p2)
   assert (isEqual (tstStructArray:draw_rect(rec), rec))

   local p3 = {x=15,y=60}
   local tri = tstStructArray:create_triangle(p1,p2,p3)
   assert (isEqual (tstStructArray:draw_triangle(tri),tri))

   local a = tstStructArray:swap({x=16,y=7})
   assert (isEqual(a,{x=7,y=16}))

   write("Arrays OK\n")


   write("---------------------\n")
   local tstEnum = tst:getTstEnum()
   assert (tstEnum:tomorrow("sun") == "mon")
   assert (tstEnum:yesterday("sun") == "sat")

   assert (tstEnum:tomorrow(4) == "fri")
   assert (tstEnum:yesterday(3) == "tue")

   write("Enums OK\n")

   local tstException = tst:getTstException()
   tstException.exception_handler = tstException_exception_handler

   write("---------------------\n")
   assert (tstException:tst_point({x=16,y=7}, {p1={x=0,y=0},p2={x=100,y=100}}))
   assert (tstException:tst_point({x=16,y=7}, {p1={x=0,y=0},p2={x=10,y=10}}) == "exception")


   assert (tstException:sqrt(49) == 7)
   assert (tstException:sqrt(-2) == "exception")
   assert (format("%1.4f",tstException:sqrt(2)) == format("%1.4f",sqrt(2)))

   write("Exceptions OK\n")


   write("---------------------\n")
   local tstUnion = tst:getTstUnion()

   local a_union = tstUnion:getField()
   assert(a_union.union_discriminator == 8)
   assert(a_union.union_value == "DEFAULT VALUE")
   assert(a_union.union_field == "m_string")
   write("Union Fields OK\n")

   tstUnion:setField("m_short",200)
   assert(tstUnion:getField().union_value == 200)
   write("Union short 1 OK\n")

   tstUnion:setField("m_short",{union_field = "m_short", union_value = 200})
   assert(tstUnion:getField().union_value == 200)
   write("Union short 2 OK\n")

   tstUnion:setField("m_ushort",{union_field = "m_ushort", union_value = 200})
   assert(tstUnion:getField().union_value == 200)
   write("Union ushort OK\n")

   tstUnion:setField("m_long",{union_field = "m_long", union_value = 150000})
   assert(tstUnion:getField().union_value == 150000)
   write("Union long OK\n")

   tstUnion:setField("m_ulong",{union_field = "m_ulong", union_value = 150000})
   assert(tstUnion:getField().union_value == 150000)
   write("Union ulong OK\n")

   tstUnion:setField("m_float",{union_field = "m_float", union_value = 150.34})
   assert(format("%3.2f",tstUnion:getField().union_value) == format("%3.2f",150.34))
   write("Union float OK\n")

   tstUnion:setField("m_double",{union_field = "m_double", union_value = 3788880000})
   assert(format("%g",tstUnion:getField().union_value) == format("%g",3788880000))
   write("Union double OK\n")

   tstUnion:setField("m_octet",{union_field = "m_octet", union_value = 8})
   assert(tstUnion:getField().union_value == 8)
   write("Union octet OK\n")

   tstUnion:setField("m_char","e")
   assert(tstUnion:getField().union_value == "e")
   write("Union char OK\n")

   tstUnion:setField("m_char",{union_field = "m_char", union_value = "a"})
   assert(tstUnion:getField().union_value == "a")
   write("Union char 2 OK\n")

   tstUnion:setField("m_string",{union_field = "m_string", union_value = "hello"})
   assert(tstUnion:getField().union_value == "hello")
   write("Union string OK\n")

   local value = lo_createtypecode("tk_string")
   tstUnion:setField("m_typecode",{union_field = "m_typecode", union_value = value})
   assert(tstUnion:getField().union_value:kind() == "tk_string")
   write("Union typecode OK\n")

   local value = "mon"
   tstUnion:setField("day",{union_field = "day", union_value = value})
   assert(tstUnion:getField().union_value == value)
   write("Union enum OK\n")

   local value = {title="T2", author= "A2"}
   tstUnion:setField("the_book",{union_field = "the_book", union_value = value})
   assert(isEqual(tstUnion:getField().union_value, value))
   write("Union struct 1 OK\n")

   local value = {x=10,y=20}
   tstUnion:setField("a_point",{union_field = "a_point", union_value = value})
   assert(isEqual(tstUnion:getField().union_value, value))
   write("Union struct 2 OK\n")

   tstUnion:setField("a_rect",{union_field = "a_rect", union_value = rec})
   assert(isEqual(tstUnion:getField().union_value, rec))
   write("Union struct 3 OK\n")

   tstUnion:setField("books",{union_field = "books", union_value = booklist})
   assert(isEqual(tstUnion:getField().union_value, booklist))
   write("Union sequence OK\n")

   tstUnion:setField("points",{union_field = "points", union_value = tri})
   assert(isEqual(tstUnion:getField().union_value, tri))
   write("Union array OK\n")

   tstUnion:setField("m_boolean",{union_field = "m_boolean", union_value = 1})
   assert(tstUnion:getField().union_value == 1)
   write("Union boolean OK\n")

   tstUnion:setField("m_any",{union_field = "m_any", union_value = "Anything"})
   assert(tstUnion:getField().union_value == "Anything")
   write("Union any OK\n")



   write("---------------------\n")
   local tstAny = tst:getTstAny()

   tstAny:setField("",200)
   assert(tstAny:getField() == 200)
   write("Any number OK\n")

   tstAny:setField("","hello")
   assert(tstAny:getField() == "hello")
   write("Any string 1 OK\n")

   tstAny:setField("",{typecode=lo_createtypecode("tk_string"),value="hello"})
   assert(tstAny:getField() == "hello")
   write("Any string 2 (typecode) OK\n")

   --tstAny:setField("",tstAttrib)
   --assert(tstAny:getField().domain == "inf.puc-rio.br")
   --write("Any proxy OK\n")

   write("---------------------\n")
   write("Test Finished\n")
   writeto()
end

if not LuaOrbCfg.getIFR() or LuaOrbCfg.getIFR():_non_existent() then
   error("The Interface Repository is not available.")
end


include("console.lua")

tst = lo_createproxy(readIOR("server.ref"),"Tst_All::Tst")

do_test()
exit()
