CORBA_2_3 = 1

function do_test(filename)
  if type(filename) == "string" then writeto(filename) end

  --
  -- Test "primitive" typecodes",
  --
  write("\nTesting Primitive Typecodes:")
  write("\n===========================\n")

  local ptypes
  if CORBA_2_3 then
  ptypes = { "tk_null", "tk_void", "tk_short", "tk_long",
	     "tk_ushort", "tk_ulong", "tk_float", "tk_double",
	     "tk_boolean", "tk_char", "tk_octet", "tk_any",
	     "tk_TypeCode", "tk_Principal", "tk_objref", "tk_longlong",
	     "tk_ulonglong", "tk_longdouble", "tk_wchar", "tk_string",
	     "tk_wstring"}
  else
  ptypes = { "tk_null", "tk_void", "tk_short", "tk_long",
	     "tk_ushort", "tk_ulong", "tk_float", "tk_double",
	     "tk_boolean", "tk_char", "tk_octet", "tk_any",
	     "tk_TypeCode", "tk_Principal", "tk_objref", "tk_longlong",
	     "tk_ulonglong", "tk_string"}
  end
  local ptcodes = {}
  foreachi(ptypes, function(i,type)
		     %ptcodes[i] = lo_createtypecode(type)
                   end)
  foreachi(ptcodes, function(i,tc)
                      write(tc:kind())
                      assert(tc:kind() == %ptypes[i])
		      write(" OK!\n")
		    end)
  assert( ptcodes[5]:equal(ptcodes[5]))
  assert( not ptcodes[5]:equal(ptcodes[4]))
  write(" --> equal ok!\n")

  if CORBA_2_3 then
    assert( ptcodes[5]:equivalent(ptcodes[5]))
    assert( not ptcodes[5]:equivalent(ptcodes[4]))
    write(" --> equivalent ok!\n")
  end

  write("\nTesting Dynamically Created TypeCodes:")
  write("\n=====================================\n")

  --
  -- Test bounded strings
  --
  local bstrtc = lo_createtypecode("tk_string",10)
  assert(bstrtc:kind() == "tk_string" and bstrtc:length() == 10)
  write("tk_string (10) OK!\n")

  if CORBA_2_3 then
    local bwstrtc = lo_createtypecode("tk_wstring",{bound = 15})
    assert(bwstrtc:kind() == "tk_wstring" and bwstrtc:length() == 15)
    write("tk_wstring (15) OK!\n")
  else
    local bwstrtc = lo_createtypecode("tk_string",{bound = 15})
    assert(bwstrtc:kind() == "tk_string" and bwstrtc:length() == 15)
    write("tk_string (15) OK!\n")
  end

  --
  -- Test interface (objref)
  --
  write("\n")
  local inttc = lo_createtypecode("tk_objref",
				  {id = "IDL:MyInt:1.0", name = "MyInt"})
  assert(inttc:kind() == "tk_objref" and inttc:id() == "IDL:MyInt:1.0"
	 and inttc:name() == "MyInt")
  local inttc2 = lo_createtypecode("tk_objref",
				  {id = "IDL:MyInt2:1.0"})
  assert(inttc2:kind() == "tk_objref" and inttc2:id() == "IDL:MyInt2:1.0"
	 and inttc2:name() == "")
  assert(inttc:equal(inttc))

  if CORBA_2_3 then
    assert(not inttc:equivalent(inttc2))
  end

  write("tk_objref (interface) OK!\n")

  --
  -- Test sequence 
  --
  write("\n")
  local seqtc = lo_createtypecode("tk_sequence", {bound = 5, element_type=bstrtc})
  assert(seqtc:kind() == "tk_sequence" and seqtc:length() == 5
	 and seqtc:content_type():kind() == "tk_string")
  write("tk_sequence OK!\n")

  --
  -- Test array
  --
  write("\n")
  local arraytc = lo_createtypecode("tk_array", {length = 5, element_type=bstrtc})
  assert(arraytc:kind() == "tk_array" and arraytc:length() == 5
	 and arraytc:content_type():kind() == "tk_string")
  assert(not arraytc:equal(seqtc))
  write("tk_array OK!\n")

  if CORBA_2_3 then
    --
    -- Test fixed
    --
    write("\n")
    local fixtc = lo_createtypecode("tk_fixed", {digits = 5, scale=8})
    assert(fixtc:kind() == "tk_fixed" and fixtc:fixed_digits() == 5
	   and fixtc:fixed_scale() == 8)
    write("tk_fixed OK!\n")
  end

  --
  -- Test alias
  --
  write("\n")
  local alitc = lo_createtypecode("tk_alias",
				  {id = "IDL:MyAlias:1.0", name = "MyAlias",
				   original_type=bstrtc})
  assert(alitc:kind() == "tk_alias" and alitc:id() == "IDL:MyAlias:1.0"
	 and alitc:name() == "MyAlias" 
         and alitc:content_type():kind() == "tk_string")

  if CORBA_2_3 then
    assert(alitc:equivalent(bstrtc))
  end
  write("tk_alias OK!\n")

  --
  -- Test Enum
  --
  write("\n")
  local enumtc = lo_createtypecode("tk_enum",
				   {id = "IDL:MyEnum:1.0",
				    members={"um","dois","tres"}})
  assert(enumtc:kind() == "tk_enum" and enumtc:member_count() == 3)
  assert(enumtc:member_name(0) == "um" and
         enumtc:member_name(1) == "dois" and
         enumtc:member_name(2) == "tres")
  write("tk_enum OK!\n")

  --
  -- Test Struct
  --
  write("\n")
  local members = {{name="string",type=lo_createtypecode("tk_string")},
                   {name="enum",type=enumtc},
		   {name="obj",type=inttc}}
  if CORBA_2_3 then
    members[4] = {name = "rec", type = lo_createtypecode("tk_recursive",
				         {id = "IDL:MyStruct:1.0"})}
  else
    members[4] = {name = "rec", 
                  type = lo_createtypecode("tk_recursive_sequence",
                                           {bound = 5, offset = 1})}
  end
  local structtc = lo_createtypecode("tk_struct",{id="IDL:MyStruct:1.0",
                 members = members})
  assert(structtc:kind() == "tk_struct" and structtc:member_count() == 4)
  assert(structtc:member_name(0) == "string" and 
	  structtc:member_type(0):kind() == "tk_string")
  assert(structtc:member_name(1) == "enum" and 
	  structtc:member_type(1):kind() == "tk_enum")
  assert(structtc:member_name(2) == "obj" and 
	  structtc:member_type(2):kind() == "tk_objref")
  if CORBA_2_3 then
    assert(structtc:member_name(3) == "rec" and 
	  structtc:member_type(3):kind() == "tk_struct" and
	  structtc:member_type(3):member_count() == 4)
  else
    assert(structtc:member_name(3) == "rec" and 
	   structtc:member_type(3):kind() == "tk_sequence" and
	   structtc:member_type(3):length() == 5 and
	   structtc:member_type(3):content_type():kind() == "tk_struct")
  end

  write("tk_struct OK!\n")

  --
  -- Test Exception
  --
  write("\n")
  local exctc = lo_createtypecode("tk_except",{id="IDL:MyException:1.0",
                             members = {{name="string",type=bstrtc}}})
  assert(exctc:kind() == "tk_except" and exctc:member_count() == 1)
  assert(exctc:member_name(0) == "string" and 
	  exctc:member_type(0):kind() == "tk_string")
  write("tk_except OK!\n")

  --
  -- Test Union
  --
  write("\n")
  local union1tc = lo_createtypecode("tk_union",
			       {id = "IDL:MyUnion:1.0",
			        discriminator_type=enumtc,
			        members={{name="string",type=bstrtc,label=0},
				  {name="obj",type=inttc,label="dois"},
				  {name="alias",type=alitc,label="default"}}})
  assert(union1tc:kind() == "tk_union" and union1tc:member_count() == 3
         and union1tc:default_index() == 2 
         and union1tc:discriminator_type():kind() == "tk_enum")
  assert(union1tc:member_name(0) == "string" and 
	  union1tc:member_type(0):kind() == "tk_string" and
          union1tc:member_label(0) == "um")
  assert(union1tc:member_name(1) == "obj" and 
	  union1tc:member_type(1):kind() == "tk_objref" and
          union1tc:member_label(1) == "dois")
  assert(union1tc:member_name(2) == "alias" and 
	  union1tc:member_type(2):kind() == "tk_alias" and
          union1tc:member_label(2) == "default")
         
                      
  write("tk_union (discriminator enum) OK!\n")

  local union2tc = lo_createtypecode("tk_union",
			     {id = "IDL:MyUnion2:1.0",
			      discriminator_type=lo_createtypecode("tk_boolean"),
			      members={{name="string",type=bstrtc,label="TRUE"},
				{name="obj",type=inttc,label="false"}}})
  assert(union2tc:kind() == "tk_union" and union2tc:member_count() == 2
         and union2tc:default_index() == -1
         and union2tc:discriminator_type():kind() == "tk_boolean")
  assert(union2tc:member_name(0) == "string" and 
	  union2tc:member_type(0):kind() == "tk_string" and
          union2tc:member_label(0))
  assert(union2tc:member_name(1) == "obj" and 
	  union2tc:member_type(1):kind() == "tk_objref" and
          not union2tc:member_label(1))

  write("tk_union (discriminator boolean/strings) OK!\n")

  local union3tc = lo_createtypecode("tk_union",
			     {id = "IDL:MyUnion3:1.0",
			      discriminator_type=lo_createtypecode("tk_boolean"),
			      members={{name="string",type=bstrtc,label=1},
				{name="obj",type=inttc}}})
  assert(union3tc:kind() == "tk_union" and union3tc:member_count() == 2
         and union3tc:default_index() == -1
         and union3tc:discriminator_type():kind() == "tk_boolean")
  assert(union3tc:member_name(0) == "string" and 
	  union3tc:member_type(0):kind() == "tk_string" and
          union3tc:member_label(0))
  assert(union3tc:member_name(1) == "obj" and 
	  union3tc:member_type(1):kind() == "tk_objref" and
          not union3tc:member_label(1))

  write("tk_union (discriminator boolean/boolean) OK!\n")

  -------------------------------------------------------------------------
  --  CORBA 2.3
  -------------------------------------------------------------------------
  if not CORBA_2_3 then
    write("\n** TypeCodes ok (CORBA 2.2) !!! **\n\n")
    return
  end

  --
  -- Test Native
  --
  write("\n")
  local nattc = lo_createtypecode("tk_native",
				  {id = "IDL:MyNative:1.0"})
  assert(nattc:kind() == "tk_native" and nattc:id() == "IDL:MyNative:1.0")
  write("tk_native OK!\n")

  --
  -- Test Abstract Interface
  --
  write("\n")
  local ainttc = lo_createtypecode("tk_abstract_interface",
				  {id = "IDL:MyAbstract:1.0"})
  assert(ainttc:kind() == "tk_abstract_interface" 
         and ainttc:id() == "IDL:MyAbstract:1.0")
  write("tk_abstract_interface OK!\n")

  --
  -- Test Value Box
  --
  write("\n")

  local vbtc = lo_createtypecode("tk_value_box", {id="IDL:MyVB:1.0",
                                            boxed_type=bstrtc})
  assert(vbtc:kind() == "tk_value_box" and vbtc:id() == "IDL:MyVB:1.0"
          and vbtc:content_type():kind() == "tk_string")
  write("tk_value_box OK!\n")

  --
  -- Test Value 
  --
  write("\n")

  local cbasetc = lo_createtypecode("tk_objref",{id="IDL:MyConcrect:1.0"})
  local vctc = lo_createtypecode("tk_value",
                           {id="IDL:ValueCustom:1.0",
 			    type_modifier="VM_CUSTOM",
			    concrete_base=cbasetc,
                            members={
                              {name="m1",type=bstrtc,access="PRIVATE_MEMBER"},
 		              {name="m2",type=inttc,access="PUBLIC_MEMBER"}}})

  assert(vctc:kind() == "tk_value" and vctc:member_count() == 2
         and vctc:type_modifier() == "VM_CUSTOM"
         and vctc:concrete_base_type():kind() == "tk_objref")
  assert(vctc:member_name(0) == "m1" and 
	  vctc:member_type(0):kind() == "tk_string" and
          vctc:member_visibility(0) == "PRIVATE_MEMBER")
  assert(vctc:member_name(1) == "m2" and 
	  vctc:member_type(1):kind() == "tk_objref" and
          vctc:member_visibility(1) == "PUBLIC_MEMBER")
  write("tk_value (custom) OK!\n")

  -- spec example
  local rectc = lo_createtypecode("tk_recursive",{id="IDL:V:1.0"})
  local vntc = lo_createtypecode("tk_value",
                           {id="IDL:V:1.0", name = "V",
 			    type_modifier="VM_NONE",
                            members={
                              {name="member",
                               type=rectc,access="PUBLIC_MEMBER"}}})

  assert(vntc:kind() == "tk_value" and vntc:member_count() == 1
         and vntc:type_modifier() == "VM_NONE"
         and vntc:concrete_base_type() == nil)
  assert(vntc:member_name(0) == "member" and 
	  vntc:member_type(0):kind() == "tk_value" and
          vntc:member_visibility(0) == "PUBLIC_MEMBER")
  write("tk_value (spec) OK!\n")

 -- ALL Tests ok!
 write("\n** TypeCodes ok (CORBA 2.3) !!! **\n\n")
 return
end

do_test()
exit()
