if not LuaOrbCfg.getIFR() then
   LuaOrbCfg.setIFR(readIOR("ir.ref"))
end

----------------------------------------------------------------------
-- Function to manage the Interface Repository
----------------------------------------------------------------------

function irfeed(file)
   execute("irfeed -ORBrepository " .. LuaOrbCfg.getIFR():_get_ior() ..  " " .. file)
end

function irdel(name)
   execute("irdel -ORBrepository " .. LuaOrbCfg.getIFR():_get_ior() .. " " .. name)
end

function mico_irfeed(file)
 execute("mico-idl -I/home/j/rcerq/braque/corba/mico/include --feed-ir --feed-included-defs -ORBIfaceRepoAddr inet:kobke.tecgraf.puc-rio.br:8888 " .. file)
end

function set_irep(file)
   LuaOrbCfg.setIFR(readIOR(file))
end

function register_Repository()
   print("\nPlease enter the ORB installation path:")
   local IDLdir = read() .. "/idl"
   local IFRfile = IDLdir .. "/OB/IFR.idl"
   print("\nPlease wait... I'm registering the Repository Interface...\n")
   execute("irfeed -I" .. IDLdir .. " -ORBrepository " .. LuaOrbCfg.getIFR():_get_ior() .. " " .. IFRfile)
end

function register_interface(interface, idl_file)
   if Rep_registered == nil then
      write("\nIs CORBA::Repository registered at the Interface Repository? ")
      local registered = read()
      if registered == "no" or registered == "n" then
         register_Repository()
      end
      Rep_registered = true
   end
   local def = LuaOrbCfg.getIFR():lookup(interface)
   if not def then
      irfeed(idl_file)
   end
end

-----------------------------------------------------------------------
-- Functions to help the construction of testing scripts
-----------------------------------------------------------------------

function table2str(table)
   local result = "{\n"
   for i,v in table do
      if type(i) == "string" then
         result = result .. "   " .. i .. " = "
      else
         result = result .. "   [" .. i .. "] = "
      end
      if type(v) == "table" then
         result = result .. table2str(v)
      elseif type(v) == "number" then
         result = result .. v
      else
         result = result .. "[[" .. v .."]]"
      end
      i,v = next(table,i)
      if i then
         result = result .. ",\n"
      else
         result = result .. "}"
      end
   end
   return result
end

function exception_handler(self, op, ex_name, ex_descr)
   print("----------------------")
   print("Exception: "..ex_name)
   print("Operation: "..op)
   local i,v = next(ex_descr, nil)
   if i then
      print(table2str(ex_descr))
   end
   print("----------------------")
   return nil
end

function isEqual(t1, t2)
   if type(t1) ~= type(t2) then
      return nil
   end
   
   if type(t1) == "table" then
      return not foreach(t1, function (i,v)
                                if not isEqual(v,%t2[i]) then return 1 end
                             end)
   else             
      return t1 == t2
   end
end          
       
function isSimilar(t1, t2)
   if type(t1) ~= type(t2) then
      return nil
   end

   if type(t1) == "table" then
      local i = 1
      local n1 = getn(t1)
      local n2 = getn(t2)
      local exists = nil
      while i <= n1 do
         local j = 1
         while j <= n2 do
            if isEqual(t1[i],t2[j]) then exists = 1 end
            j = j+ 1
         end
         if not exists then return nil else exists = nil end
         i = i + 1
      end            
      return 1
   end
end     
      
function wait(t)
   local t0 = clock()
   while (clock() - t0) <= t do end
end

----------------------------------------------------------------------
-- Auxiliary Functions to browse the Interface Repository
----------------------------------------------------------------------

function dump_info(obj)
   print(obj.id)
   print(obj.name)
   print(obj.absolute_name)
   print(obj.version)
end

function print_element(elem)
   print(elem.absolute_name)
end

function iteract_elements(list, f)
   local i = 1
   while i <= list.n do
      f(list[i])
      i = i + 1
   end
end

function list_contents( container, dk, f )
   local method = container.contents
   if not method then
      container = lo_narrow(container)
   end
   iteract_elements( container:contents(dk, nil), f )
end

function print_interfaces( container )
   list_contents(container, "dk_Interface", print_element)
end

function print_all(container)
   list_contents(container, "dk_all", print_element)
end

function print_modules(container)
   list_contents(container, "dk_Module", print_element)
end

function print_attributes(container)
   list_contents(container, "dk_Attribute", print_element)
end

function print_constants(container)
   list_contents(container, "dk_Constant", print_element)
end

function print_exceptions(container)
   list_contents(container, "dk_Exception", print_element)
end

function print_operations(container)
   list_contents(container, "dk_Operation", print_element)
end

function print_typedefs(container)
   list_contents(container, "dk_Typedef", print_element)
end

function print_alias(container)
   list_contents(container, "dk_Alias", print_element)
end

function print_structs(container)
   list_contents(container, "dk_Struct", print_element)
end

function print_unions(container)
   list_contents(container, "dk_Union", print_element)
end

function print_enums(container)
   list_contents(container, "dk_Enum", print_element)
end

function writeTypeCode(tc)
   local str = tc:name()
   if not str then
      str = tc:kind()
   end
   write(str)
end

function descrAttr(attr)
   write("   ")
   if attr.mode == 1 or attr.mode == "ATTR_READONLY" then
      write("readonly ")
   end
   write("attribute ")
   writeTypeCode(attr.type)
   write(" ")
   write(attr.name)
   write("\n")
end

function descrParam(param)
   local mode = param.mode
   if mode == 0 or mode == "PARAM_IN" then
      write("in")
   elseif mode == 1 or mode == "PARAM_OUT" then
      write("out")
   else
      write("inout")
   end
   write(" ")
   writeTypeCode(param.type)
   write(" ")
   write(param.name)
end

function descrOp(op)
   write("   ")
   if op.mode == 1 or op.mode == "OP_ONEWAY" then
      write("oneway ")
   end
   writeTypeCode(op.result)
   write(" ")
   write(op.name)
   write(" ")
   write("(")
   local i = 1
   while i <= op.parameters.n do
      descrParam(op.parameters[i])
      if i < op.parameters.n then
      write(", ")
      end
      i = i + 1
   end
   write(");")
   write("\n")
end

function Interface(interface)
   if type(interface) == "string" then
      local interface_name = interface
      interface = LuaOrbCfg.getIFR():lookup(interface_name)
      if not interface then
         interface = LuaOrbCfg.getIFR():lookup_id(interface_name)
      end
   end
   if type(interface) == "table" then
      interface = lo_narrow(interface)
      if interface:_get_interface_name() ~= "CORBA::InterfaceDef" and
         interface:_get_interface_name() ~= "::CORBA::InterfaceDef" and
         interface:_get_interface_name() ~= "InterfaceDef" then
         error("Invalid Parameter")
      end
     
      local descr = interface:describe_interface()
      write("--------------------------\n")
      write("interface " .. descr.name .. "\n")
      if descr.base_interfaces.n > 0 then
         write("   inherits from ")
         local i = 1
         while i <= descr.base_interfaces.n do
            write(descr.base_interfaces[i] .. " ")
            i = i + 1
         end
         write("\n")
      end
      write("{\n")
      foreach(descr.attributes,function(i,v) if i ~= "n" then descrAttr(v) end end)
      foreach(descr.operations,function(i,v) if i ~= "n" then descrOp(v) end end)
      write("--------------------------\n")
   end
end
