#include <stdio.h>

extern "C"
{
#include <lua.h>
#include <lualib.h>
}

#include <OB/CORBA.h>

#include <luaorb.h>

#include "kreactor-ob3.h"

#include "utils.h"
#include "cotimer.h"


// If not defined, set Lua stack size to default (1024)

#ifndef LUA_STACSIZE
#define LUA_STACKSIZE   0
#endif

CORBA_ORB_ptr orb = NULL;
CORBA_BOA_ptr boa = NULL;


static KReactor* reactor = 0;

static int my_exit(lua_State *L)
{
   boa->deactivate_impl(CORBA_ImplementationDef::_nil());
   return 0;
}

int main(int argc, char** argv)
{
   lua_State *L = lua_open(LUA_STACKSIZE);
   lua_baselibopen(L);
   lua_iolibopen(L);
   lua_strlibopen(L);
   lua_mathlibopen(L);

   orb = CORBA_ORB_init(argc, argv);
   boa = orb->BOA_init(argc, argv);
   orb -> conc_model(CORBA_ORB::ConcModelReactive);
   boa -> conc_model(CORBA_BOA::ConcModelReactive);
   boa -> init_servers();

   luaorb_open(L,orb,boa);

   cotimer_open(L);

   utils_open(L);
   int mode = utils_proc_params(L,argc,argv);

   if (mode & LO_SERVER_MODE)
   {
      if (mode & LO_INTERACTIVE_MODE)
      {
         printf("%s  %s\n> ", LUAORB_VERSION, LUAORB_COPYRIGHT);
         fflush(stdout);                                      
         reactor = new KReactor(L);                              
      }                                                       

      lua_register(L,"exit",my_exit);                           

      boa->impl_is_ready(CORBA_ImplementationDef::_nil());    

      if (reactor)                                            
         delete reactor;                                      
   }                                                          

   luaorb_close(L);
   lua_close(L);

   CORBA_release(boa);
   CORBA_release(orb);

   return 0;
}
