#include <stdio.h>
#include <iostream>
#include <unistd.h>
#include <termio.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/ioctl.h>

using namespace std;

extern "C" {
#include <lua.h>
#include <lualib.h>
}

#ifdef ORB_TAO
#include <tao/corba.h>
#else
#include <CORBA.h>
#endif

#include <luaorb.h>

#include "utils.h"

// If not defined, set Lua stack size to default (1024)

#ifndef LUA_STACSIZE
#define LUA_STACKSIZE   0
#endif

int main( int argc, char *argv[] )
{
   struct termio ctl;
   ioctl(STDIN_FILENO, TCGETA, &ctl);
   ctl.c_lflag &= ~(ICANON | ECHO);
   ioctl(STDIN_FILENO, TCSETA, &ctl);

   lua_State *L = lua_open(LUA_STACKSIZE);
   lua_baselibopen(L);
   lua_iolibopen(L);
   lua_strlibopen(L);
   lua_mathlibopen(L);

   // ORB initialization
   CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);
   
   CORBA::Object_var obj = orb -> resolve_initial_references("RootPOA");
   PortableServer::POA_var poa = PortableServer::POA::_narrow(obj);

   PortableServer::POAManager_var manager = poa->the_POAManager();
   manager->activate();

   luaorb_open(L,orb,poa);

   utils_open(L);
   int mode = utils_proc_params(L,argc,argv);

   if (mode & LO_INTERACTIVE_MODE)
   {
      char input[500];
      int  len = 0;
      cout << LUAORB_VERSION << "  " << LUAORB_COPYRIGHT << endl;
      cout << "> " << flush;
   
      while (1)
      {
         struct timeval timeout;
         fd_set fdset;
         timeout.tv_sec = 0;
         timeout.tv_usec = 10000;  
         FD_SET(STDIN_FILENO, &fdset);
         if (select(1, &fdset, 0, 0, &timeout)>0)
         {
            char c = getchar();
            switch(c)
            {
            case '\b':
            case 127:
            {
               if(len > 0)
               {
                  input[len - 1] = '\0';
                  len--;
                  cout << "\b \b" << flush;
               }
               break;
            }
            case '\r':
            case '\n':
            {
               cout << endl << flush;
               lua_dostring(L,input);
               input[0] = '\0';
               len = 0;
               cout << "> " <<  flush;
               break;
            }
            default:
            {
               if(isprint(c))
               {
                  cout << c << flush;
                  input[len] = c;
                  len++;
                  input[len] = '\0';
               }
               break;
            }
            }
         }
         if (orb->work_pending())
            orb->perform_work();
      }
   }
   else if (mode & LO_SERVER_MODE)
     orb->run();

   luaorb_close(L);
   lua_close(L);
 
   orb->destroy();
   
   return 0;
}
