#include <stdio.h>
#include <conio.h>
#include <ctype.h>

extern "C"
{
#include <lua.h>
#include <lualib.h>
}

#include <corba.h>
#include <dispatch/dispatch.h>

#include <luaorb.h>

#include "utils.h"
#include "cotimer.h"

// If not defined, set Lua stack size to default (1024)

#ifndef LUA_STACSIZE
#define LUA_STACKSIZE   0
#endif

CORBA::ORB_ptr orb = NULL;
CORBA::BOA_ptr boa = NULL;

class KHandler : public IOHandler
{
   char input_[1000];
   int len;
   lua_State* L;

   void Register()
   {
      Dispatcher::instance().startTimer(0,1000,this);
   }
public:

   KHandler(lua_State* l)
   {
      L = l;
      input_[0] = '\0';
      len = 0;
      fflush(stdin);
      setvbuf(stdin,NULL,_IONBF,0);
      Register();
   }

   virtual void timerExpired(long,long)
   {
      if(_kbhit())
      {
         char c;
         c = _getch();

         switch(c)
         {
         case '\b':
         {
            if(len > 0)
            {
               input_[len - 1] = '\0';
               len--;
               cout << "\b \b" << flush;
            }
            break;
         }
         case '\r':
         {
            cout << endl << flush;
            lua_dostring(L,input_);
            input_[0] = '\0';
            len = 0;
            cout << "> " <<  flush;
            break;
         }
         default:
         {
            if(isprint(c))
            {
               cout << c << flush;
               input_[len] = c;
               len++;
               input_[len] = '\0';
            }  
            break;
         }
         }
      }
      Register();
   }
};

static void my_exit(lua_State*L)
{
   orb->shutdown();
   return 0;
}

int main(int argc, char** argv)
{
   lua_State *L = lua_open(LUA_STACKSIZE);
   lua_baselibopen(L);
   lua_iolibopen(L);
   lua_strlibopen(L);
   lua_mathlibopen(L);

   orb = CORBA::ORB_init(argc, argv);
   boa = orb->BOA_init(argc, argv);

   luaorb_open(L,orb,boa);

   cotimer_open(L);
   utils_open(L);
   int mode = utils_proc_params(L,argc,argv);

   if (mode & LO_SERVER_MODE)
   {
      KHandler* handler = 0;

      if (mode & LO_INTERACTIVE_MODE)
      {
         handler = new KHandler(L);
         cout << LUAORB_VERSION << "  " << LUAORB_COPYRIGHT << endl;
         cout << "> " << flush;
      }

      lua_register(L,"exit",my_exit);

      boa->impl_is_ready();

      if (handler)
         delete handler;
   }

   luaorb_close(L);
   lua_close(L);

   return 0;
}
