#include <stdio.h>

extern "C"
{
#include <lua.h>
#include <lualib.h>
}

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/Logger.h>
#include <luaorb.h>

#include "wreactor-ob4.h"
#include "utils.h"

// If not defined, set Lua stack size to default (1024)

#ifndef LUA_STACSIZE
#define LUA_STACKSIZE   0
#endif

CORBA::ORB_ptr orb = NULL;
PortableServer::POA_ptr poa = NULL;


static int my_exit(lua_State* L)
{
   orb->shutdown(0);
   luaorb_close(L);
   WConsoleReactor::Close();
   return 0;
}

int main(int argc, char** argv)
{
   lua_State *L = lua_open(LUA_STACKSIZE);
   lua_baselibopen(L);
   lua_iolibopen(L);
   lua_strlibopen(L);
   lua_mathlibopen(L);

   OB::Reactor_var reactor = WConsoleReactor::Init(L,LUAORB_VERSION);

   OB::Properties_var props = OB::Properties::getDefaultProperties();
   OB::ParseArgs(argc, argv, props);
   props->setProperty("ooc.orb.conc_model", "reactive");
   props->setProperty("ooc.orb.oa.conc_model", "reactive");

   orb = OBCORBA::ORB_init(argc, argv, OB::Properties::_nil(),OB::Logger::_nil(),reactor);
   
   CORBA::Object_var obj = orb -> resolve_initial_references("RootPOA");
   PortableServer::POA_ptr poa = PortableServer::POA::_narrow(obj);
      
   OBCORBA::ORB_var obOrb = OBCORBA::ORB::_narrow(orb);

   #if OB_INTEGER_VERSION < 4000100L
   obOrb->run_init();
   #endif

   PortableServer::POAManager_var manager = poa->the_POAManager();
   manager->activate();

   luaorb_open(L,orb,poa);
   
   utils_open(L);
   int mode = utils_proc_params(L,argc,argv);

   if (mode & LO_SERVER_MODE)
   {
      if (mode & LO_INTERACTIVE_MODE)
      {
         printf("%s  %s\n> ", LUAORB_VERSION, LUAORB_COPYRIGHT);
         fflush(stdout);
         WConsoleReactor::Start();
      }

      lua_register(L,"new_exit",my_exit);
      lua_dostring(L,"original_exit = exit; exit = new_exit");

      orb->run();
   }
   
   lua_close(L);

   CORBA::release(poa);
   orb->destroy();

   return 0;
}
