#include <OB/CORBA.h>
#include <OB/EventHandlerBase.h>
#include <OB/Reactor.h>
#include <ctype.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>

extern "C"{
#include <lua.h>
};

#ifdef HAVE_TERMIO_H
#   include <termio.h>
#else
#   ifndef HAVE_TERMIOS_H
#       error Neither termio.h nor termios.h is available!
#   endif
#   include <termios.h>
#endif
#include <sys/ioctl.h>


#include "kreactor-ob4.h"

#ifdef HAVE_TERMIO_H
static struct termio TerminalSettings;
#else
static struct termios TerminalSettings;
#endif


static void ResetTerm(int sig)
{
    //
    // Reset terminal
    //
#ifdef HAVE_TERMIO_H
    ioctl(STDIN_FILENO, TCSETA, &TerminalSettings);
#else
    ioctl(STDIN_FILENO, TCSETS, &TerminalSettings);
#endif

    //
    // Send signal again
    //
    signal(sig, SIG_DFL);
    kill(getpid(), sig);
}


KReactor::KReactor(lua_State *L, OBCORBA::ORB_ptr obOrb) : input_(CORBA::string_dup(""))
{
    // Store Lua State
    myluaState = L;
    
    //
    // Register with the reactor
    //
    this->reactor = obOrb -> reactor();
    this->reactor->register_handler(this, OB::EventRead, 0, 0);
    fflush(stdin);

    //setvbuf(stdin,NULL,_IONBF,0);

    //
    // Set terminal mode
    //
#ifdef HAVE_TERMIO_H
    struct termio ctl;
    ioctl(STDIN_FILENO, TCGETA, &TerminalSettings);
    ioctl(STDIN_FILENO, TCGETA, &ctl);
#else
    struct termios ctl;
    ioctl(STDIN_FILENO, TCGETS, &TerminalSettings);
    ioctl(STDIN_FILENO, TCGETS, &ctl);
#endif
    ctl.c_lflag &= ~(ICANON | ECHO);
    ctl.c_cc[VMIN] = 1;
    ctl.c_cc[VTIME] = 0;
#ifdef HAVE_TERMIO_H
    ioctl(STDIN_FILENO, TCSETA, &ctl);
#else
    ioctl(STDIN_FILENO, TCSETS, &ctl);
#endif
    //
    // Set signal handler
    //
    signal(SIGHUP, ResetTerm);
    signal(SIGINT, ResetTerm);
    signal(SIGTERM, ResetTerm);
    signal(SIGQUIT, ResetTerm);
}

KReactor::~KReactor()
{
    //
    // Reset terminal
    //
#ifdef HAVE_TERMIO_H
    ioctl(STDIN_FILENO, TCSETA, &TerminalSettings);
#else
    ioctl(STDIN_FILENO, TCSETS, &TerminalSettings);
#endif

}

// ----------------------------------------------------------------------
// ORBacus specific event handling
// ----------------------------------------------------------------------
 
void KReactor::handleStop()
{
   this->reactor->unregister_handler(this);
}

OB::Mask KReactor::handleEvent(OB::Mask mask)
{
    assert(mask == OB::EventRead);    

    char c;
    cin.read(&c, 1);

    switch(c)
    {
    case '\b':
    case 127:
    {
        if(strlen(input_) > 0)
        {
            input_[strlen(input_) - 1] = '\0';
            cout << "\b \b" << flush;
        }
        break;
    }
    case '\n':
    {
        cout << c << flush;
        CORBA::String_var s = input_;        
        input_[0] = '\0';
        lua_dostring(myluaState, s);
        cout << "> " <<  flush;
        break;
    }
    default:
    {
        if(isprint(c))
        {
            cout << c << flush;
            input_ += c;
        }  
        break;
    }
    }
    
    return OB::EventRead;
}

