#ifndef WREACTOR_H
#define WREACTOR_H

#include <OB/EventHandlerBase_fwd.h>

#include <OB/NativeTypes.h>
#include <OB/Object.h>
#include <OB/Reactor.h>
#include <OB/list.h>

extern "C" {
#include <lua.h>
}


//
// The WConsoleReactor class
//
class WConsoleReactor : public OB::Reactor OB_COMMA_RECURSIVE_MUTEX
{
    //
    // Hide copy-constructor and assignment operator
    //
    WConsoleReactor(const WConsoleReactor&);
    void operator=(const WConsoleReactor&);

    //
    // If set to true, interrupt event dispatching immediately
    //
    bool interrupt_;

    //
    // The window used for message notifications
    //
    static HWND hWnd_;
    static unsigned long thread_handle_;

    //
    // Struct for an event handler info
    //
    struct HandlerInfo
    {
       OB::EventHandlerBase_var handler;
       OB::Handle handle;
       OB::Mask mask;
       OB::TypeMask typeMask;
    };

    //
    // Our event handler info list
    //
    OB_STL::list<HandlerInfo> handlerInfoList_;

    //
    // Return the TypeMask for the registered event handlers
    //
    OB::TypeMask evalTypeMasks() const;

    lua_State* lState;
public:
    
    static OB::Reactor_ptr Init(lua_State* state,const char* title);
    static void Start();
    static void Close();

    WConsoleReactor(lua_State*,HINSTANCE);
    virtual ~WConsoleReactor();

    virtual void register_handler(OB::EventHandlerBase*, OB::Mask, OB::TypeMask, OB::Handle);
    virtual void unregister_handler(OB::EventHandlerBase*);
    virtual bool dispatch(OB::TypeMask);
    virtual void interrupt_dispatch();
    virtual bool dispatch_one_event(CORBA::Long);
    virtual bool event_ready();

    //
    // Windows message processing function
    //
    LRESULT wndProc(HWND, UINT, WPARAM, LPARAM);
};

#endif
