//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: lo_dsi.h,v 1.16.2.2 2001/12/04 00:56:36 rcerq Exp $
//
//*****************************************************************************


#ifndef __LO_DSI_H
#define __LO_DSI_H

#include "lo_class.h"
#include "luawrap.h"

#ifdef CORBA_2_3
class LuaDSI: public PortableServer::DynamicImplementation,
              public PortableServer::RefCountServantBase
#else
class LuaDSI: public CORBA::DynamicImplementation
#endif
{
public:

   LuaOrbClass* myclass;

   static void open(lua_State *L);
   static void close(lua_State *L);

   static LuaDSI* CreateImplicitServant(lua_State *L, int pos, char* rep_id);
   static LuaDSI* Create(lua_State *L, int pos, CORBA::InterfaceDef_ptr intf, 
			 char*name=0);
   static LuaDSI* Create(lua_State *L, int pos, const char* InterfaceName, 
			 char*name=0);

   LuaDSI( lua_State *L, int pos, LuaOrbClass* lo_class, char* name = 0 );
   ~LuaDSI();

// redefined methods for interface CORBA::DynamicImplementation 
#ifdef ORB_TAO
   virtual CORBA::Boolean _is_a(const char *name,
		CORBA::Environment& ACE_TRY_ENV = TAO_default_environment());
   virtual void invoke(CORBA::ServerRequest_ptr request,
		CORBA::Environment& ACE_TRY_ENV = TAO_default_environment());
   virtual CORBA::InterfaceDef_ptr _get_interface(CORBA::Environment& ACE_TRY_ENV = TAO_default_environment());
#else
   virtual CORBA::Boolean _is_a(const char *name);
   virtual void invoke(CORBA::ServerRequest_ptr request);
   virtual CORBA::InterfaceDef_ptr _get_interface();
#endif

#if defined(CORBA_2_3) && !defined(ORB_TAO)
   virtual CORBA::RepositoryId _primary_interface(const PortableServer::ObjectId& oid, PortableServer::POA_ptr poa);
   virtual PortableServer::POA_ptr _default_POA();
#elif defined(ORB_TAO)
   virtual CORBA::RepositoryId _primary_interface(const PortableServer::ObjectId& oid, PortableServer::POA_ptr poa,CORBA::Environment& ACE_TRY_ENV = TAO_default_environment());
   virtual PortableServer::POA_ptr _default_POA(CORBA::Environment& ACE_TRY_ENV = TAO_default_environment());
#endif

// methods for interface LuaDSI
   virtual void GetLuaObject(lua_State *L);
   virtual void SetLuaObject(lua_State *L, int pos);
   virtual void RaiseException(lua_State *L, CORBA::ServerRequest_ptr request,
                               MethodDescr* descr);

// Registry data
   static const char* reg_EXCEPTION_TAG;
   static const char* reg_SERVERLIST;

// Wrapper
   static const char* wrap_LuaDSI;

private:
   lua_State* myluaState;
   int EXCEPTION_TAG;

   int tableLuaRef;
   LuaDSI* nextServer;
   LuaDSI* prevServer;
#ifdef CORBA_2_3
   PortableServer::POA_ptr poa;
#endif

   static LuaDSI*  getFirstServer(lua_State *L);
   static void setFirstServer(lua_State *L, LuaDSI* server);
   static void insertServer(lua_State *L, LuaDSI* server);
   static void removeServer(lua_State *L, LuaDSI* server);
   static void destroyServerList(lua_State *L);

public:
#ifdef ORB_ORBACUS
   OB_MAKE_NARROW_IMPL(LuaDSI)
#endif
};

#endif
