//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: lo_map.cpp,v 1.10.2.2 2001/12/04 00:56:36 rcerq Exp $
//
//*****************************************************************************


#include <orb_inc.h>
#include <string.h>

extern "C"
{
#include <lua.h>
}

#include "lo_map.h"
#include "lo_proxy.h"
#include "lo_cfg.h"
#include "lo_except.h"
#include "lo_tcode.h"
#include "lo_dsi.h"
#include "lo_util.h"

#include "lo_out.h"

#if defined(ORB_ORBACUS) && !defined(CORBA_2_3)
//
// OBFixSeq< CORBA_Octet >
//
#include <OB/TemplateI.h>

static OBTypeCodeConst _tc_OBFixSeq_CORBA_Octet(
"01000000130000000c000000010000000a00000000000000"
);

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OBInfo< OBFixSeq< CORBA_Octet > >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OBInfo< OBFixSeq< CORBA_Octet > >)
#endif
#endif

void operator<<=(CORBA_Any& any, OBFixSeq< CORBA_Octet >* p)
{
    static const OBInfo< OBFixSeq< CORBA_Octet > > info;
    any.replace(_tc_OBFixSeq_CORBA_Octet, p, true, &info);
}

void operator<<=(CORBA_Any& any, const OBFixSeq< CORBA_Octet >& seq)
{
    static const OBInfo< OBFixSeq< CORBA_Octet > > info;
    OBFixSeq< CORBA_Octet >* p = new OBFixSeq< CORBA_Octet >(seq);
    any.replace(_tc_OBFixSeq_CORBA_Octet, p, true, &info);
}

CORBA_Boolean operator>>=(const CORBA_Any& any, const OBFixSeq< CORBA_Octet >*& p)
{
    if(any.check_type(_tc_OBFixSeq_CORBA_Octet))
    {
        if(!any.info())
        {
            const OBBuffer* cbufp = (const OBBuffer*)any.value();
            const CORBA_Octet* coct = cbufp -> data;
            OBFixSeq< CORBA_Octet >* v = new OBFixSeq< CORBA_Octet >;
            OBUnmarshal(*v, coct, false);
            (CORBA_Any&)any <<= v;
        }

        p = (const OBFixSeq< CORBA_Octet >*)any.value();
        return true;
    }
    else
        return false;
}
#endif


class LuaMapping
{
public:
   static DYNANY(DynAny_ptr) createDynAny(LuaOrbCfg* loCfg, CORBA::TypeCode_ptr tc);
   static DYNANY(DynAny_ptr) createDynAny(LuaOrbCfg* loCfg, CORBA::Any* any);

   static void fromLua(lua_State *L, int pos, CORBA::TypeCode_ptr tc,
                       DYNANY(DynAny_ptr) dynAny);
   static void toLua(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny);

   static void toShort(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toUShort(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toLong(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toULong(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toFloat(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toDouble(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toString(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toBoolean(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toChar(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toOctet(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toTypeCode(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toObjRef(lua_State *L, int pos, CORBA::TypeCode_ptr tc, 
                        DYNANY(DynAny_ptr) dynAny);
   static void toStruct(lua_State *L, int pos, CORBA::TypeCode_ptr tc,
                        DYNANY(DynAny_ptr) dynAny);
   static void toSequence(lua_State *L, int pos, CORBA::TypeCode_ptr tc,
                          DYNANY(DynAny_ptr) dynAny);
   static void toArray(lua_State *L, int pos, CORBA::TypeCode_ptr tc, 
                       DYNANY(DynAny_ptr) dynAny);
   static void toAny(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny);
   static void toUnion(lua_State *L, int pos, CORBA::TypeCode_ptr,
                       DYNANY(DynAny_ptr) dynAny);
   static void toEnum(lua_State *L, int pos, CORBA::TypeCode_ptr tc,
                      DYNANY(DynAny_ptr) dynAny);

   static void fromShort(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromLong(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromUShort(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromULong(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromFloat(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromDouble(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromString(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromBoolean(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromChar(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromOctet(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromObjRef(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny);
   static void fromStruct(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny);
   static void fromSequence(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny);
   static void fromArray(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny);
   static void fromAny(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromUnion(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny);
   static void fromEnum(lua_State *L, DYNANY(DynAny_ptr) dynAny);
   static void fromTypeCode(lua_State *L, DYNANY(DynAny_ptr) dynAny);

};

DYNANY(DynAny_ptr) LuaMapping::createDynAny(LuaOrbCfg* loCfg, CORBA::TypeCode_ptr tc)
{
   DYNANY(DynAny_ptr) dynAny = DYNANY(DynAny)::_nil();

   if (tc->kind() == CORBA(tk_alias))
   {
      CORBA::TypeCode_var tc_v = tc->content_type();
      return LuaMapping::createDynAny(loCfg, tc_v);
   }

#ifdef CORBA_2_3
   dynAny = loCfg->orb_dynany_factory->create_dyn_any_from_type_code(tc);
#else
   switch (tc->kind())
   {
      case CORBA(tk_alias):
      case CORBA(tk_short):
      case CORBA(tk_long):
      case CORBA(tk_ushort):
      case CORBA(tk_ulong):
      case CORBA(tk_float):
      case CORBA(tk_double):
      case CORBA(tk_string):
      case CORBA(tk_boolean):
      case CORBA(tk_char):
      case CORBA(tk_octet):
      case CORBA(tk_objref):
      case CORBA(tk_any):
      case CORBA(tk_TypeCode):
         dynAny = loCfg->orb->create_basic_dyn_any(tc);
         break;
      case CORBA(tk_struct):
      case CORBA(tk_except):
         dynAny = loCfg->orb->create_dyn_struct(tc);
         break;
      case CORBA(tk_union):
         dynAny = loCfg->orb->create_dyn_union(tc);
         break;
      case CORBA(tk_enum):
         dynAny = loCfg->orb->create_dyn_enum(tc);
         break;
      case CORBA(tk_sequence):
         dynAny = loCfg->orb->create_dyn_sequence(tc);
         break;
      case CORBA(tk_array):
         dynAny = loCfg->orb->create_dyn_array(tc);
         break;
      case CORBA(tk_Principal):
         throw Err_NotImpl("Principal");
         break;
      default:
         throw Err_NotImpl("Unknown Type");
         break;
   }
#endif
   return dynAny;
}


DYNANY(DynAny_ptr) LuaMapping::createDynAny(LuaOrbCfg* loCfg, CORBA::Any* any)
{
   DYNANY(DynAny_ptr) dynAny = DYNANY(DynAny)::_nil();
#ifdef CORBA_2_3
   dynAny = loCfg->orb_dynany_factory->create_dyn_any(*any);
#else
   dynAny = loCfg->orb->create_dyn_any(*any);
#endif
   return dynAny;
}

void LuaMapping::toShort(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   if (!lua_isnumber(L, pos))
       throw Err_Param("short");
   short luaval = (short)lua_tonumber(L, pos);
   dynAny->insert_short(luaval);
}

void LuaMapping::toUShort(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   if (!lua_isnumber(L, pos))
      throw Err_Param("unsigned short");
   unsigned short luaval = (unsigned short)lua_tonumber(L, pos);
   dynAny->insert_ushort(luaval);
}

void LuaMapping::toLong(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   if (!lua_isnumber(L, pos))
      throw Err_Param("long");
   long luaval = (long)lua_tonumber(L, pos);
   dynAny->insert_long(luaval);
}

void LuaMapping::toULong(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   if (!lua_isnumber(L, pos))
      throw Err_Param("unsigned long");
   unsigned long luaval = (unsigned long)lua_tonumber(L, pos);
   dynAny->insert_ulong(luaval);
}

void LuaMapping::toFloat(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   if (!lua_isnumber(L, pos))
      throw Err_Param("float");
   float luaval = (float)lua_tonumber(L, pos);
   dynAny->insert_float(luaval);
}

void LuaMapping::toDouble(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   if (!lua_isnumber(L, pos))
      throw Err_Param("double");
   double luaval = (double)lua_tonumber(L, pos);
   dynAny->insert_double(luaval);
}

void LuaMapping::toString(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   const char* luaval = lua_tostring(L, pos);
   if (luaval == NULL)
      throw Err_Param("string");
   dynAny->insert_string(luaval);
}

void LuaMapping::toBoolean(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   CORBA::Boolean b;
   if (lua_isnil(L, pos))
       b = 0;
   else 
       b = 1;
   dynAny->insert_boolean(b);
}

void LuaMapping::toChar(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   if (!lua_isstring(L, pos))
      throw Err_Param("char");
   char* str = (char*) lua_tostring(L,pos);
   dynAny->insert_char(str[0]);
}

void LuaMapping::toOctet(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   if (!lua_isnumber(L, pos))
      throw Err_Param("octet");
   CORBA::Octet luaval = (CORBA::Octet)lua_tonumber(L, pos);
   dynAny->insert_octet(luaval);
}

void LuaMapping::toTypeCode(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   LuaTypeCode* ltc = LuaTypeCode::from_lua(L, pos);

   if (ltc == NULL)
      throw Err_Param("typecode");

   dynAny->insert_typecode(ltc->get_tc());
}

void LuaMapping::toObjRef(lua_State *L, int pos, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   try {
      CORBA::Object_var obj;
      LuaProxy* proxy = LuaProxy::from_lua(L, pos);

      if (proxy != NULL)
         obj = CORBA::Object::_duplicate(proxy->orb_object);
      else if (LuaOrbUtil::isIndexable(L, pos))
      {
         LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
         LuaDSI* luadsi = (LuaDSI*) Wrapper->from_lua(L, pos);
         if (luadsi == NULL)
         {
            CORBA::String_var name = tc->name();
            CORBA::String_var id = tc->id();
            luadsi = LuaDSI::CreateImplicitServant(L, pos, id);
            if (luadsi==NULL)
               throw Err_Param("object reference");
            
         }
#ifdef CORBA_2_3
         PortableServer::POA_var poa = luadsi->_default_POA();
         obj = poa->servant_to_reference(luadsi);
#else
         obj = CORBA::Object::_duplicate(luadsi);
#endif
      }
      else if (lua_isnil(L, pos))
         obj = CORBA::Object::_nil();
      else if (lua_isstring(L, pos))
      {
         try {
            const char* ior = lua_tostring(L, pos);
            obj = (LuaOrbCfg::getConfiguration(L))->orb->string_to_object(ior);
         }
         catch (CORBA::Exception& excep) {
            LO_OUT(L)  << excep << endl;
            throw Err_Param("object reference");
         }
      }
      else
         throw Err_Param("object reference");

      dynAny->insert_reference(obj);
   }
   catch(CORBA::Exception& excep) {
      LO_OUT(L) << excep << endl;
      throw;
   }
   catch(...) {
      throw;
   }
}

void LuaMapping::toStruct(lua_State *L, int pos, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   CORBA::TypeCode_var tc_v;
   DYNANY(DynAny_var) component;
   CORBA::String_var member_name;

   if (!LuaOrbUtil::isIndexable(L, pos))
      throw Err_Param("struct");

   for (unsigned long i =0; i < tc->member_count(); i++)
   {
      member_name = tc->member_name(i);
      lua_pushstring(L, member_name);
      lua_gettable(L, pos);

      try {
         component = dynAny->current_component();
         tc_v = tc->member_type(i);
         LuaMapping::fromLua(L, lua_gettop(L), tc_v, component);
         dynAny->next();
      }
      catch (Err_Param&) {
         throw Err_Param("struct");
      }
      lua_pop(L, 1);
   }
}

void LuaMapping::toSequence(lua_State *L, int pos, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   unsigned long i, length;
   unsigned long max_len = tc->length();
   CORBA::TypeCode_var c_type = tc->content_type();

   if (c_type->kind() == CORBA(tk_octet) && lua_isstring(L, pos))
   {
      const char* buffer = lua_tostring(L, pos);
      length = (unsigned long) lua_strlen(L, pos);
      if (max_len > 0 && length > max_len)
         length = max_len;

      CORBA::Octet* octet_buf = CORBA::OctetSeq::allocbuf(length);
      memcpy(octet_buf,buffer,length);
      CORBA::OctetSeq seq(length,length,octet_buf,1);
      CORBA::Any anyValue;
      anyValue <<= seq;
      dynAny->from_any(anyValue);
   }
   else if (LuaOrbUtil::isIndexable(L, pos))
   {
      if (lua_istable(L, pos))
          length = (unsigned long) lua_getn(L, pos);
      else
      {
          lua_pushstring(L, "n");
          lua_gettable(L, pos);
          if (lua_isnumber(L, -1))
              length = (unsigned long) lua_tonumber(L, -1);
          else
             throw Err_Param("sequence");
          lua_pop(L, 1);
      }
      if (max_len > 0 && length > max_len)
         length = max_len;

      if (c_type->kind() == CORBA(tk_octet))
      {
#if 0
         DYNANY(DynSequence_var) dyn_seq = DYNANY(DynSequence)::_narrow(dynAny);
#ifdef CORBA_2_3
         dyn_seq->set_length(length);
#else
         dyn_seq->length(length);
#endif
         for (i = 1; i <= length; i++)
         {
            lua_pushnumber(L, (double)i);
            lua_gettable(L, pos);
            dynAny->insert_octet((CORBA::Octet) lua_tonumber(L, -1));
#ifdef CORBA_2_3
            dynAny->next();
#endif
            lua_pop(L, 1);
         }
#else
         CORBA::Octet* octet_buf = CORBA::OctetSeq::allocbuf(length);
         for (i = 1; i <= length; i++)
         {
            lua_pushnumber(L, (double)i);
            lua_gettable(L, pos);
            octet_buf[i-1] = (CORBA::Octet) lua_tonumber(L, -1);
            lua_pop(L, 1);
         }

         CORBA::OctetSeq seq(length,length,octet_buf,1);
         CORBA::Any anyValue;
         anyValue <<= seq;
         dynAny->from_any(anyValue);
#endif
      }
      else
      {
         DYNANY(DynSequence_var) dyn_seq = DYNANY(DynSequence)::_narrow(dynAny);
#ifdef CORBA_2_3
         dyn_seq->set_length(length);
#else
         dyn_seq->length(length);
#endif
         for (i = 1; i <= length; i++)
         {
            lua_pushnumber(L, (double)i);
            lua_gettable(L, pos);
            DYNANY(DynAny_var) component = dyn_seq->current_component();
            LuaMapping::fromLua(L, lua_gettop(L), c_type, component);
            dyn_seq->next();
            lua_pop(L, 1);
         }
      }
   }
   else
      throw Err_Param("sequence");
}

void LuaMapping::toArray(lua_State *L, int pos, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   unsigned long i;
   unsigned long len = tc->length();
   CORBA::TypeCode_var c_type = tc->content_type();

   if (!LuaOrbUtil::isIndexable(L, pos))
      throw Err_Param("array");

   for (i = 0; i < len; i++)
   {
      lua_pushnumber(L, (double)i+1);
      lua_gettable(L, pos);

      try {
         DYNANY(DynAny_var) component = dynAny->current_component();
         LuaMapping::fromLua(L, lua_gettop(L),  c_type, component);
         dynAny->next();
      }
      catch (Err_Param&) {
         throw Err_Param("array");
      }
      lua_pop(L, 1);
   }
}

void LuaMapping::toAny(lua_State *L, int pos, DYNANY(DynAny_ptr) dynAny)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);
   DYNANY(DynAny_ptr) aux_dynAny = DYNANY(DynAny)::_nil();

   if (lua_isnumber(L, pos)) //map to double
   {
      aux_dynAny = LuaMapping::createDynAny(loCfg, CORBA(_tc_double));
      LuaMapping::toDouble(L, pos, aux_dynAny);
   }
   else if (lua_isstring(L, pos)) //map to string
   {
      aux_dynAny = LuaMapping::createDynAny(loCfg, CORBA(_tc_string));
      LuaMapping::toString(L, pos, aux_dynAny);
   }
   else if (lua_istable(L, pos))
   {
      LuaWrapper* Wrapper = LuaWrapper::getWrapper(L, LuaDSI::wrap_LuaDSI);
      if (LuaProxy::isProxy(L, pos) || Wrapper->from_lua(L, pos))
      {
         aux_dynAny = LuaMapping::createDynAny(loCfg, CORBA(_tc_Object));
         LuaMapping::toObjRef(L, pos, CORBA(_tc_Object), aux_dynAny);
      }
      else if (LuaTypeCode::isTypeCode(L,pos))
      {
         aux_dynAny = LuaMapping::createDynAny(loCfg, CORBA(_tc_TypeCode));
         LuaMapping::toTypeCode(L, pos, aux_dynAny);
      }
      else
      {
         lua_pushstring(L,"typecode");
         lua_gettable(L, pos);
         int top = lua_gettop(L);

         if ( !lua_isnil(L, -1)  &&
              lua_istable(L, -1) && LuaTypeCode::isTypeCode(L, top) )
         {
            LuaTypeCode* ltc = LuaTypeCode::from_lua(L, top);
            aux_dynAny = LuaMapping::createDynAny(loCfg, ltc->get_tc());
            lua_pushstring(L, "value");
            lua_gettable(L, pos);
            LuaMapping::fromLua(L, lua_gettop(L), ltc->get_tc(), aux_dynAny);
            lua_pop(L, 2);
         }
         else
         {
            lua_pop(L, 1); // pop typecode
            //map to sequence of Any
            CORBA::TypeCode_var tc = loCfg->orb->create_sequence_tc(0,
                                        CORBA(_tc_any));
            aux_dynAny = LuaMapping::createDynAny(loCfg, tc);
            LuaMapping::fromLua(L, pos, tc, aux_dynAny);
         }
      }
   }
   else
      throw Err_Param("CORBA::Any");

   CORBA::Any_var any = aux_dynAny->to_any();
#ifdef ORB_TAO
   dynAny->insert_any(any.in());
#else
   dynAny->insert_any(any);
#endif
   aux_dynAny->destroy();
   CORBA::release(aux_dynAny);
}

void LuaMapping::toUnion(lua_State *L, int pos, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   unsigned long i;
   DYNANY(DynUnion_var) dyn_union = DYNANY(DynUnion)::_narrow(dynAny);
   int err_flag = 0;
   CORBA::TypeCode_var tc_v;

#ifdef CORBA_2_3
   DYNANY(DynAny_var) discriminator = dyn_union->get_discriminator();
#else
   DYNANY(DynAny_var) discriminator = dyn_union->discriminator();
#endif
   DYNANY(DynAny_var) member;

   if (lua_istable(L, pos))
   {
      lua_pushstring(L, "union_field");
      lua_gettable(L, pos);
      if ( !lua_isnil(L, -1) && lua_isstring(L, -1))  
      {
         const char* name = lua_tostring(L, -1);
         CORBA::String_var member_name;
         for (i = 0; i < tc->member_count(); i++)
         {
            member_name = tc->member_name(i);
            if (strcmp(name, member_name) == 0)
            {
               if (tc->default_index() == (long)i)
#ifdef CORBA_2_3
                  dyn_union->set_to_default_member();
#else
                  dyn_union->set_as_default(1);
#endif
               else
               {
                  CORBA::Any_var any_v = tc->member_label(i);
#ifdef ORB_TAO
                  discriminator->from_any(any_v.in());
#else
                  discriminator->from_any(any_v);
#endif
               }
               member = dyn_union->member();
               tc_v = tc->member_type(i);

               lua_pushstring(L, "union_value");
               lua_gettable(L, pos);
               LuaMapping::fromLua(L, lua_gettop(L), tc_v, member);

               lua_pop(L, 2); // pop union_field and union_value
               return;
            }
         }
         lua_pop(L, 1); // pop union_field
      }
      else
      { 
         lua_pushstring(L,"union_discriminator");
         lua_gettable(L, pos);
         if (!lua_isnil(L, -1))
         {
            tc_v = tc->discriminator_type();
            LuaMapping::fromLua(L, lua_gettop(L), tc_v, discriminator);
#ifdef CORBA_2_3
            if (dyn_union->component_count() == 2)
#else
            dyn_union->rewind();
            if (dyn_union->next())
#endif
            {
                member = dyn_union->member();
                tc_v = member->type();

                lua_pushstring(L, "union_value");
                lua_gettable(L, pos);
                LuaMapping::fromLua(L, lua_gettop(L), tc_v, member);
                lua_pop(L, 3); // pop union_field, union_discriminator
                //  and union_value
            }
            else
               lua_pop(L, 2); // pop union_field and union_discriminator
            return;
         }
         lua_pop(L, 2);       // pop union_field and union_discriminator
      }
   }

   //if it's not a union table then to try matching
   for (i =0; i < tc->member_count(); i++)
   {
      err_flag = 0;
      CORBA::Any_var any_v = tc->member_label(i);
#ifdef ORB_TAO
      discriminator->from_any(any_v.in());
#else
      discriminator->from_any(any_v);
#endif
      try
      {
         member = dyn_union->member();
         tc_v = tc->member_type(i);
         LuaMapping::fromLua(L, pos, tc_v, member);
      }
      catch(Err_Param&)
      {
         err_flag = 1;
      }
      if (err_flag == 0)
         break;
   }

   if (err_flag != 0)
      throw Err_Param("union");
}

void LuaMapping::toEnum(lua_State *L, int pos, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   DYNANY(DynEnum_var) dyn_enum = DYNANY(DynEnum)::_narrow(dynAny);

   if (lua_isnumber(L, pos))
   {
      unsigned long val = (unsigned long)lua_tonumber(L, pos);
      if ( val < tc->member_count() )
      {
#ifdef CORBA_2_3
         dyn_enum->set_as_ulong(val);
#else
         dyn_enum->value_as_ulong(val);
#endif
         return;
      }
   }
   else if (lua_isstring(L, pos))
   {
      const char* member = lua_tostring(L, pos);
      unsigned long i;
      CORBA::String_var member_name;
      for (i = 0; i < tc->member_count(); i++)
      {
         member_name = tc->member_name(i);
         if (strcmp(member,member_name) == 0)
         {
#ifdef CORBA_2_3
            dyn_enum->set_as_string(member);
#else
            dyn_enum->value_as_string(member);
#endif
            return;
         }
      }
   }
   throw Err_Param("enum");
}

void LuaMapping::fromLua(lua_State *L, int pos, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   switch (tc->kind())
   {
   case CORBA(tk_short):
      VERBOSE_OUT(L, "tk_short")
      LuaMapping::toShort(L,pos,dynAny);
      break;
   case CORBA(tk_long):
      VERBOSE_OUT(L, "tk_long")
      LuaMapping::toLong(L,pos,dynAny);
      break;
   case CORBA(tk_ushort):
      VERBOSE_OUT(L, "tk_ushort")
      LuaMapping::toUShort(L,pos,dynAny);
      break;
   case CORBA(tk_ulong):
      VERBOSE_OUT(L, "tk_ulong")
      LuaMapping::toULong(L,pos,dynAny);
      break;
   case CORBA(tk_float):
      VERBOSE_OUT(L, "tk_float")
      LuaMapping::toFloat(L,pos,dynAny);
      break;
   case CORBA(tk_double):
      VERBOSE_OUT(L, "tk_double")
      LuaMapping::toDouble(L,pos,dynAny);
      break;
   case CORBA(tk_string):
      VERBOSE_OUT(L, "tk_string")
      LuaMapping::toString(L,pos,dynAny);
      break;
   case CORBA(tk_boolean):
      VERBOSE_OUT(L, "tk_boolean")
      LuaMapping::toBoolean(L,pos,dynAny);
      break;
   case CORBA(tk_char):
      VERBOSE_OUT(L, "tk_char")
      LuaMapping::toChar(L,pos,dynAny);
      break;
   case CORBA(tk_octet):
      VERBOSE_OUT(L, "tk_octet")
      LuaMapping::toOctet(L,pos,dynAny);
      break;
   case CORBA(tk_objref):
      VERBOSE_OUT(L, "tk_objref")
      LuaMapping::toObjRef(L,pos,tc,dynAny);
      break;
   case CORBA(tk_struct):
      VERBOSE_OUT(L, "tk_struct")
      LuaMapping::toStruct(L,pos,tc,dynAny);
      break;
   case CORBA(tk_any):
      VERBOSE_OUT(L, "tk_any")
      LuaMapping::toAny(L,pos,dynAny);
      break;
   case CORBA(tk_union):
      VERBOSE_OUT(L, "tk_union")
      LuaMapping::toUnion(L,pos,tc,dynAny);
      break;
   case CORBA(tk_enum):
      VERBOSE_OUT(L, "tk_enum")
      LuaMapping::toEnum(L,pos,tc,dynAny);
      break;
   case CORBA(tk_sequence):
      VERBOSE_OUT(L, "tk_sequence")
      LuaMapping::toSequence(L,pos,tc,dynAny);
      break;
   case CORBA(tk_array):
      VERBOSE_OUT(L, "tk_array")
      LuaMapping::toArray(L,pos,tc,dynAny);
      break;
   case CORBA(tk_alias):
   {
      VERBOSE_OUT(L, "tk_alias")
      CORBA::TypeCode_var tc_v = tc->content_type();
      LuaMapping::fromLua(L,pos,tc_v,dynAny);
      break;
   }
   case CORBA(tk_except):
      VERBOSE_OUT(L, "tk_except")
      LuaMapping::toStruct(L,pos,tc,dynAny);
      break;
   case CORBA(tk_TypeCode):
      VERBOSE_OUT(L, "tk_TypeCode")
      LuaMapping::toTypeCode(L,pos,dynAny);
      break;
   default:
      break;
   }
}

//
//  lua2any receives the index (stack position) of the Lua object
//

CORBA::Any* lua2any(lua_State *L, int pos, CORBA::TypeCode_ptr tc)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);
   DYNANY(DynAny_ptr) dynAny = LuaMapping::createDynAny(loCfg,tc);

   LuaMapping::fromLua(L, pos, tc, dynAny);

   CORBA::Any* anyValue = dynAny->to_any();
   dynAny->destroy();
   CORBA::release(dynAny);
   return anyValue;
}

/********************************************************************/

void LuaMapping::fromShort(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   short ret = dynAny->get_short();
   lua_pushnumber(L, (double)ret);
}

void LuaMapping::fromLong(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   long ret = dynAny->get_long();
   lua_pushnumber(L, (double)ret);
}

void LuaMapping::fromUShort(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   unsigned short ret = dynAny->get_ushort();
   lua_pushnumber(L, (double)ret);
}

void LuaMapping::fromULong(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   unsigned long ret = dynAny->get_ulong();
   lua_pushnumber(L, (double)ret);
}

void LuaMapping::fromFloat(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   float ret = dynAny->get_float();
   lua_pushnumber(L, (double)ret);
}

void LuaMapping::fromDouble(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   double ret = dynAny->get_double();
   lua_pushnumber(L, ret);
}

void LuaMapping::fromString(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   CORBA::String_var ret = dynAny->get_string();
   lua_pushstring(L, ret);
}

void LuaMapping::fromBoolean(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   CORBA::Boolean ret = dynAny->get_boolean();
   if (ret)
      lua_pushnumber(L, (double)1.0);
   else
      lua_pushnil(L);
}

void LuaMapping::fromChar(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   CORBA::Char ret = dynAny->get_char();
   char str[2];
   str[0] = ret;
   str[1] = 0;
   lua_pushstring(L, str);
}

void LuaMapping::fromOctet(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   CORBA::Octet ret = dynAny->get_octet();
   lua_pushnumber(L, (double)ret);
}

#ifdef CORBA_2_3
static LuaDSI* getLocalServant(lua_State *L, CORBA::Object_ptr obj, PortableServer::POA_ptr poa)
{
   PortableServer::Servant servant;
   servant = poa->reference_to_servant(obj);
   VERBOSE_OUT(L, "ObjRef->Servant conversion")

   LuaDSI* l_dsi = dynamic_cast<LuaDSI*>(servant);
   if (l_dsi)
      VERBOSE_OUT(L, "Servant->LuaDSI conversion")

   servant->_remove_ref();
   return l_dsi;
}
#endif

void LuaMapping::fromObjRef(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   CORBA::Object_ptr ret = dynAny->get_reference();
   if (CORBA::is_nil(ret))
      lua_pushnil(L);
   else
   {
#ifdef CORBA_2_3
      if (LuaOrbCfg::modeConfigured(L, SERVANT_OPTIMIZATION))
      {
         LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);
         try {
            LuaDSI* l_dsi = getLocalServant(L,ret,loCfg->poa);
            if (l_dsi)
            {
               l_dsi->GetLuaObject(L);
               return;
            }
         }
         catch(CORBA::UserException& excep)
         {
            VERBOSE_OUT2(L, "Servant optimization failed for Transient POA: ", excep)
         }
         try {
            LuaDSI* l_dsi = getLocalServant(L,ret,loCfg->persistent_poa);
            if (l_dsi)
            {
               l_dsi->GetLuaObject(L);
               return;
            }
         }
         catch(CORBA::UserException& excep)
         {
            VERBOSE_OUT2(L, "Servant optimization failed for Persistent POA: ", excep)
         }
      }
#endif
      CORBA::String_var id = tc->id();
      LuaProxy::push(L, new LuaProxy(L,id,ret));
   }
}

void LuaMapping::fromStruct(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   lua_newtable(L);
   for (unsigned long i = 0; i < tc->member_count(); i++)
   {
      DYNANY(DynAny_var) component = dynAny->current_component();
      CORBA::TypeCode_var tc_v = tc->member_type(i);
      CORBA::String_var member_name = tc->member_name(i);

      lua_pushstring(L, member_name);
      LuaMapping::toLua(L,tc_v,component);   // Lua object at the top of the stack
      lua_rawset(L, -3);
      dynAny->next();
   }
}

void LuaMapping::fromSequence(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   unsigned long i, len;
   CORBA::TypeCode_var c_type = tc->content_type();

#ifdef CORBA_2_3
   len = dynAny->component_count();
#else
   DYNANY(DynSequence_var) dyn_seq = DYNANY(DynSequence)::_narrow(dynAny);
   len = dyn_seq->length();
#endif

   if (c_type->kind() == CORBA(tk_octet))
   {
      const CORBA::OctetSeq* seq;
      CORBA::Any_var any = dynAny->to_any();
#ifdef ORB_TAO
      any.in() >>= seq;
#else
      any >>= seq;
#endif
      if (LuaOrbCfg::modeConfigured(L, OCTETSEQ_AS_STRING))
      {
#if defined(ORB_ORBACUS) && !defined(CORBA_2_3)
         const CORBA::Octet* data = seq->data();
#else
         const CORBA::Octet* data = seq->get_buffer();
#endif
         lua_pushlstring(L,(char*)data,len);
         return;
      }
      else
      {
         lua_newtable(L);
         for (i = 0; i < len; i++)
         {
            lua_pushnumber(L, (double)(*seq)[i]);
            lua_rawseti(L, -2, i+1);
         }
      }
   }
   else
   {
      lua_newtable(L);
      for (i = 0; i < len; i++)
      {
         DYNANY(DynAny_var) component = dynAny->current_component();
         LuaMapping::toLua(L,c_type,component);
         lua_rawseti(L, -2, i+1);
         dynAny->next();
      }
   }

   lua_pushstring(L,"n");
   lua_pushnumber(L, (double)len);
   lua_rawset(L, -3);
}

void LuaMapping::fromArray(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   unsigned long len = tc->length();
   CORBA::TypeCode_var c_type = tc->content_type();

   lua_newtable(L);
   for (unsigned long i = 0; i < len; i++)
   {
      DYNANY(DynAny_var) component = dynAny->current_component();
      LuaMapping::toLua(L, c_type,component);
      lua_rawseti(L, -2, i+1);
      dynAny->next();
   }
   lua_pushstring(L, "n");
   lua_pushnumber(L, (double)len);
   lua_rawset(L, -3);
}

void LuaMapping::fromAny(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);
   CORBA::Any_var any = dynAny->get_any();
   DYNANY(DynAny_ptr) aux_dynAny = LuaMapping::createDynAny(loCfg, any);
   CORBA::TypeCode_var tc_v = aux_dynAny->type();
   LuaMapping::toLua(L,tc_v, aux_dynAny);
   aux_dynAny->destroy();
   CORBA::release(aux_dynAny);
}

void LuaMapping::fromUnion(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   DYNANY(DynUnion_var) dyn_union = DYNANY(DynUnion)::_narrow(dynAny);
#ifdef CORBA_2_3
   DYNANY(DynAny_var) discriminator = dyn_union->get_discriminator();
#else
   DYNANY(DynAny_var) discriminator = dyn_union->discriminator();
#endif
   CORBA::TypeCode_var tc_v =tc->discriminator_type();

   lua_newtable(L);

   lua_pushstring(L, "union_discriminator");
   LuaMapping::toLua(L, tc_v, discriminator);
   lua_rawset(L, -3);

#ifdef CORBA_2_3
   if (dyn_union->component_count() == 2)
#else
   dyn_union->rewind();
   if (dyn_union->next())
#endif
   {
      lua_pushstring(L, "union_field");
      CORBA::String_var name = dyn_union->member_name();
      lua_pushstring(L, name);
      lua_rawset(L, -3);

      lua_pushstring(L, "union_value");
      DYNANY(DynAny_var) member = dyn_union->member();
      tc_v = member->type();
      LuaMapping::toLua(L, tc_v, member);
      lua_rawset(L, -3);
   }
}

void LuaMapping::fromEnum(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   DYNANY(DynEnum_var) dyn_enum = DYNANY(DynEnum)::_narrow(dynAny);

   if (LuaOrbCfg::modeConfigured(L, ENUM_NUMERIC_MODE))
#ifdef CORBA_2_3
      lua_pushnumber(L, (double)dyn_enum->get_as_ulong());
#else
      lua_pushnumber(L, (double)dyn_enum->value_as_ulong());
#endif
   else
   {
      CORBA::String_var str;
#ifdef CORBA_2_3
      str = dyn_enum->get_as_string();
#else
      str = dyn_enum->value_as_string();
#endif
      lua_pushstring(L, str);
   }
}

void LuaMapping::fromTypeCode(lua_State *L, DYNANY(DynAny_ptr) dynAny)
{
   CORBA::TypeCode_var tc_v = dynAny->get_typecode();
   LuaTypeCode* ltc = new LuaTypeCode(tc_v);
   LuaTypeCode::push(L, ltc);
}

void LuaMapping::toLua(lua_State *L, CORBA::TypeCode_ptr tc, DYNANY(DynAny_ptr) dynAny)
{
   switch (tc->kind())
   {
     case CORBA(tk_short):
         VERBOSE_OUT(L, "tk_short")
         LuaMapping::fromShort(L, dynAny);
         break;
     case CORBA(tk_long):
         VERBOSE_OUT(L, "tk_long")
         LuaMapping::fromLong(L, dynAny);
         break;
     case CORBA(tk_ushort):
         VERBOSE_OUT(L, "tk_ushort")
         LuaMapping::fromUShort(L, dynAny);
         break;
     case CORBA(tk_ulong):
         VERBOSE_OUT(L, "tk_ulong")
         LuaMapping::fromULong(L, dynAny);
         break;
     case CORBA(tk_float):
         VERBOSE_OUT(L, "tk_float")
         LuaMapping::fromFloat(L, dynAny);
         break;
     case CORBA(tk_double):
         VERBOSE_OUT(L, "tk_double")
         LuaMapping::fromDouble(L, dynAny);
         break;
     case CORBA(tk_string):
         VERBOSE_OUT(L, "tk_string")
         LuaMapping::fromString(L, dynAny);
         break;
     case CORBA(tk_boolean):
         VERBOSE_OUT(L, "tk_boolean")
         LuaMapping::fromBoolean(L, dynAny);
         break;
     case CORBA(tk_char):
         VERBOSE_OUT(L, "tk_char")
         LuaMapping::fromChar(L, dynAny);
         break;
     case CORBA(tk_octet):
         VERBOSE_OUT(L, "tk_octet")
         LuaMapping::fromOctet(L, dynAny);
         break;
     case CORBA(tk_any):
         VERBOSE_OUT(L, "tk_any")
         LuaMapping::fromAny(L, dynAny);
         break;
     case CORBA(tk_enum):
         VERBOSE_OUT(L, "tk_enum")
         LuaMapping::fromEnum(L, dynAny);
         break;
     case CORBA(tk_TypeCode):
         VERBOSE_OUT(L, "tk_TypeCode")
         LuaMapping::fromTypeCode(L, dynAny);
         break;
     case CORBA(tk_alias):
     {
         VERBOSE_OUT(L, "tk_alias")
         CORBA::TypeCode_var tc_v = tc->content_type();
         LuaMapping::toLua(L, tc_v, dynAny);
         break;
     }
     case CORBA(tk_objref):
         {
            VERBOSE_OUT(L, "tk_objref")
            CORBA::String_var name = tc->name();
            CORBA::String_var id = tc->id();
            VERBOSE_OUT(L, name)
            VERBOSE_OUT(L, id)
         }
         LuaMapping::fromObjRef(L, tc, dynAny);
         break;
     case CORBA(tk_except):
         VERBOSE_OUT(L, "tk_except")
         LuaMapping::fromStruct(L, tc,dynAny);
         break;
     case CORBA(tk_struct):
         VERBOSE_OUT(L, "tk_struct")
         LuaMapping::fromStruct(L, tc, dynAny);
         break;
     case CORBA(tk_sequence):
         VERBOSE_OUT(L, "tk_sequence")
         LuaMapping::fromSequence(L, tc, dynAny);
         break;
     case CORBA(tk_array):
         VERBOSE_OUT(L, "tk_array")
         LuaMapping::fromArray(L, tc, dynAny);
         break;
     case CORBA(tk_union):
         VERBOSE_OUT(L, "tk_union")
         LuaMapping::fromUnion(L, tc, dynAny);
         break;
     default:
         break;
   }
}

//
//  any2lua returns the Lua object at the top of the stack
//

void any2lua(lua_State *L, CORBA::TypeCode_ptr tc, CORBA::Any* resultAny)
{
   LuaOrbCfg* loCfg = LuaOrbCfg::getConfiguration(L);
   DYNANY(DynAny_ptr) dynAny = LuaMapping::createDynAny(loCfg, resultAny);

   LuaMapping::toLua(L, tc, dynAny); // Lua object returned at the top of the stack

   dynAny->destroy();
   CORBA::release(dynAny);
}

