//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: lo_proxy.h,v 1.30.2.1 2001/08/07 13:33:02 moura Exp $
//
//*****************************************************************************


#ifndef LO_PROXY_H
#define LO_PROXY_H

#include "lo_class.h"
#include "luawrap.h"

enum InvokeMode { NORMAL_CALL, DEFERRED_CALL};

class LuaProxy
{
   int lua_this;

   void CreateLuaThis(lua_State *L);

public:
   class DeferredClosure
   {
      int _finished;
   public:
      
      lua_State* myluaState;

      LuaProxy* object;
      CORBA::Request_ptr req;
      CORBA::Any* resultAny;
      CORBA::NVList_ptr arguments;
      MethodDescr* method_descr;
      AttribDescr* attrib_descr;

      DeferredClosure()
      { object = NULL; req = NULL; resultAny = NULL; arguments = NULL;
        method_descr = NULL; attrib_descr = NULL; _finished = 0;}
      ~DeferredClosure()
      { if (req != NULL)
          CORBA_release(req);
      }
      int Finished(lua_State *L);
      void Wait(lua_State *L);
      unsigned int GetResult(lua_State *L);
   };

   // Registry data
   static const char* reg_TAG;
   static const char* reg_UDTAG;

   // Wrapper
   static const char* wrap_DEFCLOSURE;

   static void open(lua_State *L);
   static void close(lua_State *L);
   static void push(lua_State *L, LuaProxy* obj);
   static LuaProxy* check_tag(lua_State *L, int narg, int tag);
   static LuaProxy* from_lua(lua_State *L, int pos);
   static char* parsename(char* full_name, InvokeMode* mode);
   static int isProxy(lua_State *L, int pos);

   LuaOrbClass* myclass;
   CORBA::Object_ptr orb_object;
   int locked;

   LuaProxy(lua_State *L, const char* interfacename, CORBA::Object_ptr orb_obj);
   virtual ~LuaProxy();

   void GetLuaThis(lua_State *L);
   void SetLuaThis(lua_State *L, int pos);
   void DiscardLuaThis(lua_State *L);

   void GetLuaExceptionHandler(lua_State *L);
   unsigned int RaiseException(lua_State *L, CORBA::Exception* excep, 
			       MethodDescr* descr);

   unsigned int Call(lua_State *L, int pos, unsigned int nparams,
                     MethodDescr* method_descr, 
		     InvokeMode mode);
   int SetAttr(lua_State *L, int pos, AttribDescr* attrib_descr, 
		InvokeMode mode);
   void GetAttr(lua_State *L, AttribDescr* attrib_descr, InvokeMode mode);
};


#endif
