//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: lo_tcode.cpp,v 1.6.2.2 2001/12/04 00:56:36 rcerq Exp $
//
//*****************************************************************************

#include <stdio.h>
#include <orb_inc.h>

extern "C"
{
#include <lua.h>
#include <lauxlib.h>
}

#include "lo_cfg.h"
#include "lo_tcode.h"
#include "lo_map.h"
#include "lo_util.h"

const char* LuaTypeCode::reg_TAG   = "LuaTypeCode_TAG";
const char* LuaTypeCode::reg_UDTAG = "LuaTypeCode_UDTAG";

typedef struct {
    const char* kind;
    CORBA(TypeCode_ptr) const_val;
} TCode_desc;

static TCode_desc  tc_desc[] = {
   {"tk_null",      CORBA(_tc_null)}, 
   {"tk_void",      CORBA(_tc_void)},  
   {"tk_short",     CORBA(_tc_short)}, 
   {"tk_long",      CORBA(_tc_long)},
   {"tk_ushort",    CORBA(_tc_ushort)}, 
   {"tk_ulong",     CORBA(_tc_ulong)},
   {"tk_float",     CORBA(_tc_float)}, 
   {"tk_double",    CORBA(_tc_double)},
   {"tk_boolean",   CORBA(_tc_boolean)}, 
   {"tk_char",      CORBA(_tc_char)},
   {"tk_octet",     CORBA(_tc_octet)}, 
   {"tk_any",       CORBA(_tc_any)},
   {"tk_TypeCode",  CORBA(_tc_TypeCode)}, 
   {"tk_Principal", CORBA(_tc_Principal)},
   {"tk_objref",    NULL},
   {"tk_struct",    NULL},
   {"tk_union",     NULL},
   {"tk_enum",      NULL},
   {"tk_string",    NULL},
   {"tk_sequence",  NULL},
   {"tk_array",     NULL},
   {"tk_alias",     NULL},
   {"tk_except",    NULL},
   {"tk_longlong",  CORBA(_tc_longlong)}, 
   {"tk_ulonglong", CORBA(_tc_ulonglong)},
   
#ifdef CORBA_2_3
   {"tk_longdouble",         CORBA(_tc_longdouble)}, 
   {"tk_wchar",              CORBA(_tc_wchar)}, 
   {"tk_wstring",            NULL},
   {"tk_fixed",              NULL},
   {"tk_value",              NULL},
   {"tk_value_box",          NULL},
   {"tk_native",             NULL},
   {"tk_abstract_interface", NULL},
#endif
   {NULL,NULL}
};

#ifdef CORBA_2_3
static const char* value_modifiers[]={"VM_NONE","VM_CUSTOM", "VM_ABSTRACT","VM_TRUNCATABLE"};
static const char* visibility[]={"PRIVATE_MEMBER","PUBLIC_MEMBER"};
#endif

class DynamicTC
{
   LuaOrbCfg* loCfg;

public:
   DynamicTC(LuaOrbCfg* cfg) { loCfg = cfg; }

   CORBA(TypeCode_ptr) create_tc(lua_State *L, CORBA(TCKind) tk);
   CORBA(TypeCode_ptr) create_objref(lua_State *L);
   CORBA(TypeCode_ptr) create_string(lua_State *L, CORBA(TCKind) tk);
   CORBA(TypeCode_ptr) create_struct( lua_State *L, CORBA(TCKind) tk);
   CORBA(TypeCode_ptr) create_union(lua_State *L);
   CORBA(TypeCode_ptr) create_enum(lua_State *L);
   CORBA(TypeCode_ptr) create_sequence(lua_State *L);
   CORBA(TypeCode_ptr) create_array(lua_State *L);
   CORBA(TypeCode_ptr) create_alias(lua_State *L);
   CORBA(TypeCode_ptr) create_fixed(lua_State *L);

#ifndef CORBA_2_3
   CORBA(TypeCode_ptr) create_recursive_sequence(lua_State *L);
#endif

#ifdef CORBA_2_3
   CORBA(TypeCode_ptr) create_recursive(lua_State *L);
   CORBA(TypeCode_ptr) create_value(lua_State *L);
   CORBA(TypeCode_ptr) create_value_box(lua_State *L);
   CORBA(TypeCode_ptr) create_interface(lua_State *L, CORBA(TCKind) tk);
#endif

private:
   static unsigned long get_NumericField(lua_State *L, int narg, const char* field);
   static CORBA(TypeCode_ptr) get_TypeCodeField(lua_State *L, int narg, const char* field);
   static const char* get_RepID(lua_State *L, int narg);
   static const char* get_Name(lua_State *L, int narg);
   static void get_Members(lua_State *L, int narg);
   static const char* get_MemberName(lua_State *L, int member);
   static CORBA(TypeCode_ptr) get_MemberType(lua_State *L, int member);
   static CORBA::Any get_MemberLabel(lua_State *L, int member,
                              CORBA(TypeCode_ptr) discriminator_type);
#ifdef CORBA_2_3
   static CORBA(ValueModifier) get_TypeModifier(lua_State *L, int narg);
   static CORBA(Visibility) get_MemberAccess(lua_State *L, int member);
#endif
};


static int ltc_equal(lua_State *L)
{
   LuaTypeCode* ltc  =  LuaTypeCode::check_tag(L, 1);
   CORBA(TypeCode_ptr) tc = ltc->get_tc();
   if (tc) {
      LuaTypeCode* ltc2 = LuaTypeCode::check_tag(L, 2);
      CORBA::Boolean ret = tc->equal(ltc2->get_tc());
      if (ret)
         lua_pushnumber(L, 1);
      else
         lua_pushnil(L);
   }
   else
      lua_pushnil(L);
  return 1;
}

#ifdef CORBA_2_3
static int ltc_equivalent(lua_State *L)
{
   LuaTypeCode* ltc  =  LuaTypeCode::check_tag(L, 1);
   CORBA(TypeCode_ptr) tc = ltc->get_tc();
   if (tc) {
      LuaTypeCode* ltc2 = LuaTypeCode::check_tag(L, 2);
      CORBA::Boolean ret = tc->equivalent(ltc2->get_tc());
      if (ret)
         lua_pushnumber(L, 1);
      else
         lua_pushnil(L);
   }
   else
      lua_pushnil(L);
  return 1;
}

#ifndef ORB_TAO
static int ltc_get_compact_typecode(lua_State *L)
{
   LuaTypeCode* ltc  =  LuaTypeCode::check_tag(L, 1);
   CORBA(TypeCode_ptr) tc = ltc->get_tc();
   if (tc)
   {
     CORBA(TypeCode_var) tc_v = tc->get_compact_typecode();
     LuaTypeCode::push(L, new LuaTypeCode(tc_v));
   }
   else
      lua_pushnil(L);
  return 1;
}
#endif
#endif

static int ltc_kind(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   CORBA(TypeCode_ptr) tc = ltc->get_tc();
   if (tc)
      lua_pushstring(L, tc_desc[tc->kind()].kind);
   else
      lua_pushnil(L);
  return 1;
}

static int ltc_id(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_objref) || tk == CORBA(tk_struct) ||
       tk == CORBA(tk_union)  || tk == CORBA(tk_enum)   ||
       tk == CORBA(tk_alias)  ||
#ifdef CORBA_2_3
       tk == CORBA(tk_value)  || tk == CORBA(tk_value_box) ||
       tk == CORBA(tk_native) || tk == CORBA(tk_abstract_interface) ||
#endif
       tk == CORBA(tk_except))
   {
      CORBA::String_var id = ltc->get_tc()->id();
      lua_pushstring(L, id);
   }
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_name(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_objref) || tk == CORBA(tk_struct) ||
       tk == CORBA(tk_union)  || tk == CORBA(tk_enum)   ||
       tk == CORBA(tk_alias)  ||
#ifdef CORBA_2_3
       tk == CORBA(tk_value)  || tk == CORBA(tk_value_box) ||
       tk == CORBA(tk_native) || tk == CORBA(tk_abstract_interface) ||
#endif
       tk == CORBA(tk_except))
   {
      CORBA::String_var name = ltc->get_tc()->name();
      lua_pushstring(L, name);
   }
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_member_count(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_struct) || tk == CORBA(tk_union) ||
       tk == CORBA(tk_enum)   || 
#ifdef CORBA_2_3
       tk == CORBA(tk_value)   || 
#endif
       tk == CORBA(tk_except))
      lua_pushnumber(L, (double) ltc->get_tc()->member_count());
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_member_name(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   unsigned long index = (unsigned long) luaL_check_number(L, 2);

   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_struct) || tk == CORBA(tk_union) ||
       tk == CORBA(tk_enum)   ||
#ifdef CORBA_2_3
       tk == CORBA(tk_value)   || 
#endif
       tk == CORBA(tk_except))
   {
      CORBA::String_var member_name = ltc->get_tc()->member_name(index);
      lua_pushstring(L, member_name);
   }
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_member_type(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   unsigned long index = (unsigned long) luaL_check_number(L, 2);

   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_struct) || tk == CORBA(tk_union) ||
#ifdef CORBA_2_3
       tk == CORBA(tk_value)   || 
#endif
       tk == CORBA(tk_except))
   {
      CORBA(TypeCode_var) tc_v = ltc->get_tc()->member_type(index);
      LuaTypeCode::push(L, new LuaTypeCode(tc_v));
   }
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_member_label(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   unsigned long index = (unsigned long) luaL_check_number(L, 2);

   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_union))
   {
      if ((long) index == ltc->get_tc()->default_index())
      {
         lua_pushstring(L, "default");
      }
      else
      {
         CORBA::Any_var any = ltc->get_tc()->member_label(index);
         any2lua(L, any->type(),any);
      }
   }
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_discriminator_type(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_union))
   {
      CORBA(TypeCode_var) tc_v = ltc->get_tc()->discriminator_type();
      LuaTypeCode::push(L, new LuaTypeCode(tc_v));
   }
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_default_index(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_union))
      lua_pushnumber(L, (double) ltc->get_tc()->default_index());
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_length(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_string) || 
#ifdef CORBA_2_3
       tk == CORBA(tk_wstring) ||
#endif
       tk == CORBA(tk_sequence) || tk == CORBA(tk_array))
      lua_pushnumber(L, (double) ltc->get_tc()->length());
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_content_type(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_sequence) || tk == CORBA(tk_array) ||
#ifdef CORBA_2_3
       tk == CORBA(tk_value_box)   || 
#endif
       tk == CORBA(tk_alias))
   {
      CORBA(TypeCode_var) tc_v = ltc->get_tc()->content_type();
      LuaTypeCode::push(L, new LuaTypeCode(tc_v));
   }
   else
      lua_pushnil(L);
   return 1;
}

#if defined(CORBA_2_3) && !defined(ORB_TAO)
static int ltc_fixed_digits(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_fixed))
      lua_pushnumber(L, (double) ltc->get_tc()->fixed_digits());
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_fixed_scale(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_fixed))
      lua_pushnumber(L, (double) ltc->get_tc()->fixed_scale());
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_member_visibility(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   unsigned long index = (unsigned long) luaL_check_number(L, 2);

   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_value))
      lua_pushstring(L, visibility[ltc->get_tc()->member_visibility(index)]);
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_type_modifier(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_value))
      lua_pushstring(L, value_modifiers[ltc->get_tc()->type_modifier()]);
   else
      lua_pushnil(L);
   return 1;
}

static int ltc_concrete_base_type(lua_State *L)
{
   LuaTypeCode* ltc = LuaTypeCode::check_tag(L, 1);
   enum CORBA(TCKind) tk = ltc->get_tc()->kind();

   if (tk == CORBA(tk_value))
   {
      CORBA::TypeCode_var tc_v = ltc->get_tc()->concrete_base_type();
      if (tc_v)
         LuaTypeCode::push(L, new LuaTypeCode(tc_v));
      else
         lua_pushnil(L);
   }
   else
      lua_pushnil(L);
   return 1;
}
#endif

static int ltc_createtypecode(lua_State *L)
{
   LuaTypeCode* ltc = NULL;

   const char *lkind = luaL_check_string(L, 1);
   unsigned long tk = 0;
   TCode_desc* dp;

   for (dp = tc_desc; dp->kind != NULL; dp++)
   {
      if (strcmp(lkind,dp->kind) == 0)
         break;
      tk++;
   }

   // Constant (built-int) Type Code ?
   if ((dp->kind != NULL) && (dp->const_val != NULL))
      ltc = new LuaTypeCode(CORBA::TypeCode::_duplicate(dp->const_val));
   
   else {
    // Dynamic TypeCode

      DynamicTC* dynTC = new DynamicTC(LuaOrbCfg::getConfiguration(L));

      if (dp->kind != NULL)
         ltc = new LuaTypeCode(dynTC->create_tc(L, (CORBA::TCKind) tk));
      else
#ifdef CORBA_2_3
         if (strcmp(lkind,"tk_recursive") == 0)
            ltc = new LuaTypeCode(dynTC->create_recursive(L));
#else
         if (strcmp(lkind,"tk_recursive_sequence") == 0)
            ltc = new LuaTypeCode(dynTC->create_recursive_sequence(L));
#endif
         else 
            luaL_argerror(L, 1,"invalid/unsupported TypeCode kind");
   }

   if (ltc != NULL)
      LuaTypeCode::push(L, ltc);
   else
      lua_pushnil(L);

   return 1;
}

static int gc_fb(lua_State *L)
{
   LuaTypeCode* ltc = (LuaTypeCode*) lua_touserdata(L, 1);
   delete ltc;
   return 0;
}

static void register_methods(lua_State *L, struct luaL_reg* list, int n)
{
   // table is at the top of the stack

   for (int i = 0; i < n; i++)
   {
      lua_pushstring(L, list[i].name);
      lua_pushcfunction(L, list[i].func);
      lua_rawset(L, -3);
   }
}

/*****************************************************************************/

static struct luaL_reg functions_tb []= {
   {"lo_createtypecode",(lua_CFunction)ltc_createtypecode}
};

static struct luaL_reg methods_tb [] ={
                           {"equal",              (lua_CFunction)ltc_equal},
#ifdef CORBA_2_3
                           {"equivalent",         (lua_CFunction)ltc_equivalent},
#ifndef ORB_TAO
                           {"get_compact_typecode", (lua_CFunction)ltc_get_compact_typecode},
#endif
#endif
                           {"kind",               (lua_CFunction)ltc_kind},
                           {"id",                 (lua_CFunction)ltc_id},
                           {"name",               (lua_CFunction)ltc_name},
                           {"member_count",       (lua_CFunction)ltc_member_count},
                           {"member_name",        (lua_CFunction)ltc_member_name},
                           {"member_type",        (lua_CFunction)ltc_member_type},
                           {"member_label",       (lua_CFunction)ltc_member_label},
                           {"discriminator_type", (lua_CFunction)ltc_discriminator_type},
                           {"default_index",      (lua_CFunction)ltc_default_index},
                           {"length",             (lua_CFunction)ltc_length},
                           {"content_type",       (lua_CFunction)ltc_content_type},
#if defined(CORBA_2_3) && !defined(ORB_TAO)
                           {"fixed_digits",       (lua_CFunction)ltc_fixed_digits},
                           {"fixed_scale",        (lua_CFunction)ltc_fixed_scale},
                           {"member_visibility",  (lua_CFunction)ltc_member_visibility},
                           {"type_modifier",      (lua_CFunction)ltc_type_modifier},
                           {"concrete_base_type", (lua_CFunction)ltc_concrete_base_type}
#endif
                        };

void LuaTypeCode::open(lua_State *L)
{
   int TAG = lua_newtag(L);
   lua_pushnumber(L, TAG);
   lua_setglobal(L, "LUATYPECODE_TAG");

   int UDTAG = lua_newtag(L);
   lua_pushcfunction(L, (lua_CFunction)gc_fb);
   lua_settagmethod(L, UDTAG,"gc");

   luaL_openlib(L, functions_tb,sizeof(functions_tb)/sizeof(struct luaL_reg));

   // TypeCode tags are saved in Lua Registry
   lua_getregistry(L);
   lua_pushstring(L, LuaTypeCode::reg_TAG);
   lua_pushnumber(L, TAG);
   lua_settable(L, -3);
   lua_pushstring(L, LuaTypeCode::reg_UDTAG);
   lua_pushnumber(L, UDTAG);
   lua_settable(L, -3);

   lua_pop(L, 1);
}

LuaTypeCode::~LuaTypeCode()
{
   if (this->corba_tc)
     CORBA_release(this->corba_tc);
}

void LuaTypeCode::push(lua_State *L, LuaTypeCode* ltc)
{
   lua_getregistry(L);
   lua_pushstring(L, LuaTypeCode::reg_TAG);
   lua_gettable(L, -2);
   int TAG = (int) lua_tonumber(L, -1);
   lua_pushstring(L, LuaTypeCode::reg_UDTAG);
   lua_gettable(L, -3);
   int UDTAG = (int) lua_tonumber(L, -1);
   lua_pop(L, 3);

   lua_newtable(L);
   lua_pushstring(L, "_CORBA_TYPECODE_REF_");
   lua_pushusertag(L, (void *) ltc, UDTAG);
   lua_rawset(L, -3);

   register_methods(L, methods_tb,
                    sizeof(methods_tb)/sizeof(struct luaL_reg));

   lua_settag(L, TAG);  // table returned at the top of the stack
}

LuaTypeCode* LuaTypeCode::check_tag(lua_State *L, int narg)
{
   LuaTypeCode* ltc = LuaTypeCode::from_lua(L, narg);
   luaL_arg_check(L, (ltc!=NULL),narg,"it isn't a CORBA::TypeCode Object");
   return ltc;
}

int LuaTypeCode::isTypeCode(lua_State *L, int pos)
{
   lua_getregistry(L);
   lua_pushstring(L, LuaTypeCode::reg_TAG);
   lua_gettable(L, -2);
   int res = (lua_tag(L, pos) == (int) lua_tonumber(L,-1));
   lua_pop(L, 2);
   return res;
}

LuaTypeCode* LuaTypeCode::from_lua(lua_State *L, int narg)
{
   if (LuaTypeCode::isTypeCode(L, narg))
   {
      lua_pushstring(L, "_CORBA_TYPECODE_REF_");
      lua_gettable(L, narg);

      LuaTypeCode* ltc = (LuaTypeCode*) lua_touserdata(L, -1);

      lua_pop(L,1);   // remove userdata from the stack
      return ltc;
   }
   else
      return NULL;
}

LuaTypeCode::LuaTypeCode( CORBA(TypeCode_ptr) tc )
{
   if (tc)
     this->corba_tc = CORBA::TypeCode::_duplicate(tc);
   else
     this->corba_tc = tc;
}

/*****************************************************************************/
/* TypeCode Creation                                                         */
/*****************************************************************************/

CORBA(TypeCode_ptr) DynamicTC::create_tc(lua_State *L, CORBA(TCKind) tk)
{
   switch (tk)
   {
      case CORBA(tk_objref):
         return this->create_objref(L);

      case CORBA(tk_string):
#ifdef CORBA_2_3
      case CORBA(tk_wstring):
#endif
         return this->create_string(L,tk);

      case CORBA(tk_struct):
      case CORBA(tk_except):
         return this->create_struct(L,tk);

      case CORBA(tk_union):
        return this->create_union(L);

      case CORBA(tk_enum):
        return this->create_enum(L);

      case CORBA(tk_sequence):
         return this->create_sequence(L);
      case CORBA(tk_array):
         return this->create_array(L);

      case CORBA(tk_alias):
         return this->create_alias(L);
         
#ifdef CORBA_2_3
      case CORBA(tk_fixed):
         return this->create_fixed(L);

      case CORBA(tk_value):
         return this->create_value(L);

      case CORBA(tk_value_box):
         return this->create_value_box(L);

      case CORBA(tk_native):
      case CORBA(tk_abstract_interface):
         return this->create_interface(L,tk);
#endif
      default:
         return NULL;
    }
}

CORBA(TypeCode_ptr) DynamicTC::create_objref(lua_State *L)
{
   if (lua_gettop(L) < 2)
      return CORBA::TypeCode::_duplicate(CORBA(_tc_Object));
   else
   {
      luaL_checktype(L, 2, LUA_TTABLE);
      const char *id = DynamicTC::get_RepID(L,2);
      const char *name  = DynamicTC::get_Name(L,2);
      return loCfg->orb->create_interface_tc(id,name);
   }

}
CORBA(TypeCode_ptr) DynamicTC::create_string(lua_State *L, CORBA(TCKind) tk)
{
   if (lua_gettop(L) < 2)
   {
      // unbounded string
#ifdef CORBA_2_3
      if ( tk == CORBA(tk_string) )
         return CORBA::TypeCode::_duplicate(CORBA(_tc_string));
      else
         return CORBA::TypeCode::_duplicate(CORBA(_tc_wstring));
#else
      return CORBA::TypeCode::_duplicate(CORBA(_tc_string));
#endif
   }

   unsigned long bound;
   if (lua_isnumber(L, 2))
      bound = (unsigned long)lua_tonumber(L, 2);
   else
   {
      if (!lua_istable(L, 2))
         luaL_argerror(L, 2,"number or table expected");
      bound = DynamicTC::get_NumericField(L, 2,"bound");
   }
#ifdef CORBA_2_3
   if ( tk == CORBA(tk_string) )
      return loCfg->orb->create_string_tc(bound);
   else 
      return loCfg->orb->create_wstring_tc(bound);
#else
   return loCfg->orb->create_string_tc(bound);
#endif
}

CORBA(TypeCode_ptr) DynamicTC::create_struct(lua_State *L, CORBA(TCKind) tk)
{
   luaL_checktype(L, 2, LUA_TTABLE);
   const char *id = DynamicTC::get_RepID(L, 2);
   const char *name  = DynamicTC::get_Name(L, 2);

   CORBA(StructMemberSeq) members;

   DynamicTC::get_Members(L, 2);  // "members" table returned at the top of the stack
   int n = lua_getn(L, -1);
   members.length((unsigned long) n);

   for (int i = 0 ; i < n ; i++)
   {
      lua_rawgeti(L, -1, (i+1));  // next member at the top of the stack 
      if (!lua_istable(L, -1))
        luaL_verror(L, "invalid member description (#%d): table expected", i+1);
      members[i].name = CORBA::string_dup(DynamicTC::get_MemberName(L, i+1));
      members[i].type = DynamicTC::get_MemberType(L,i+1);
      members[i].type_def = CORBA::IDLType::_nil();
      lua_pop(L, 1);
   }

   lua_pop(L, 1); // remove "members" table from stack

   if (tk == CORBA(tk_struct))
     return loCfg->orb->create_struct_tc(id,name,members);
   else
     return loCfg->orb->create_exception_tc(id,name,members);
}

CORBA(TypeCode_ptr) DynamicTC::create_union(lua_State *L)
{
   luaL_checktype(L, 2, LUA_TTABLE);
   const char *id = DynamicTC::get_RepID(L, 2);
   const char *name  = DynamicTC::get_Name(L, 2);
   CORBA(TypeCode_ptr) discriminator_type = DynamicTC::get_TypeCodeField(L, 2, "discriminator_type");
   CORBA(UnionMemberSeq) members;

   DynamicTC::get_Members(L, 2);  // "members" table returned at the top of the stack
   int n = lua_getn(L, -1);
   members.length((unsigned long) n);

   for (int i = 0 ; i < n ; i++)
   {
      lua_rawgeti(L, -1, (i+1));  // next member at the top of the stack 
      if (!lua_istable(L, -1))
        luaL_verror(L, "invalid member description (#%d): table expected", i+1);

      members[i].name = CORBA::string_dup(DynamicTC::get_MemberName(L, i+1));
      members[i].type = DynamicTC::get_MemberType(L, i+1);
      members[i].label = DynamicTC::get_MemberLabel(L, i+1, discriminator_type);
      members[i].type_def = CORBA::IDLType::_nil();
      lua_pop(L, 1);
   }

   lua_pop(L, 1); // remove "members" table from stack
   return loCfg->orb->create_union_tc(id,name,discriminator_type,members);
}

CORBA(TypeCode_ptr) DynamicTC::create_sequence(lua_State *L)
{
   luaL_checktype(L, 2, LUA_TTABLE);

   unsigned long bound = DynamicTC::get_NumericField(L, 2,"bound");
   CORBA(TypeCode_ptr) element_type = DynamicTC::get_TypeCodeField(L, 2, "element_type");
   return loCfg->orb->create_sequence_tc(bound,element_type);
}

CORBA(TypeCode_ptr) DynamicTC::create_array(lua_State *L)
{
   luaL_checktype(L, 2, LUA_TTABLE);
   unsigned long bound = DynamicTC::get_NumericField(L, 2, "length");
   CORBA(TypeCode_ptr) element_type = DynamicTC::get_TypeCodeField(L, 2, "element_type");
   return loCfg->orb->create_array_tc(bound,element_type);
}

CORBA(TypeCode_ptr) DynamicTC::create_alias(lua_State *L)
{
   luaL_checktype(L, 2, LUA_TTABLE);
   const char *id = DynamicTC::get_RepID(L, 2);
   const char *name  = DynamicTC::get_Name(L, 2);
   CORBA(TypeCode_ptr) original_type = DynamicTC::get_TypeCodeField(L, 2, "original_type");
   return loCfg->orb->create_alias_tc(id,name,original_type);
}

CORBA(TypeCode_ptr) DynamicTC::create_enum(lua_State *L)
{
   luaL_checktype(L, 2, LUA_TTABLE);
   const char *id = DynamicTC::get_RepID(L, 2);
   const char *name  = DynamicTC::get_Name(L, 2);

   CORBA(EnumMemberSeq) members;

   DynamicTC::get_Members(L, 2);  // "members" table returned at the top of the stack
   int n = lua_getn(L, -1);
   members.length((unsigned long) n);

   for (int i = 0 ; i < n ; i++)
   {
      lua_rawgeti(L, -1, (i+1));  // next member at the top of the stack 
      if (!lua_isstring(L, -1))
        luaL_verror(L, "invalid member (#%d)", i+1);
      members[i] = CORBA::string_dup(lua_tostring(L, -1));
      lua_pop(L, 1);
   }

   lua_pop(L, 1); // remove "members" table from stack
   return loCfg->orb->create_enum_tc(id,name,members);
}

#ifdef CORBA_2_3
CORBA(TypeCode_ptr) DynamicTC::create_fixed(lua_State *L)
{
   luaL_checktype(L, 2, LUA_TTABLE);
   unsigned short digits = (unsigned short) DynamicTC::get_NumericField(L, 2, "digits");
   unsigned short scale  = (unsigned short) DynamicTC::get_NumericField(L, 2, "scale");

   return loCfg->orb->create_fixed_tc(digits,scale);
}

CORBA(TypeCode_ptr) DynamicTC::create_value(lua_State *L)
{

   luaL_checktype(L, 2, LUA_TTABLE);
   const char *id = DynamicTC::get_RepID(L, 2);
   const char *name  = DynamicTC::get_Name(L, 2);
   CORBA(ValueModifier) type_modifier = DynamicTC::get_TypeModifier(L, 2);

   // concrete base can be nil !
   CORBA(TypeCode_ptr) concrete_base = CORBA::TypeCode::_nil();
   lua_pushstring(L, "concrete_base");
   lua_gettable(L, 2);
   if (!lua_isnil(L, -1))
   {
      LuaTypeCode* ltc = LuaTypeCode::from_lua(L, lua_gettop(L));
      if (ltc == NULL)
        lua_error(L, "'concrete_base' isn't a CORBA::TypeCode Object");
      concrete_base = ltc->get_tc();
   }
   lua_pop(L, 1);

   CORBA(ValueMemberSeq) members;

   DynamicTC::get_Members(L, 2);  // "members" table returned at the top of the stack
   int n = lua_getn(L, -1);
   members.length((unsigned long) n);

   for (int i = 0 ; i < n ; i++)
   {
      lua_rawgeti(L, -1, (i+1));  // next member at the top of the stack 
      if (!lua_istable(L, -1))
        luaL_verror(L, "invalid member description (#%d): table expected", i+1);

      members[i].name = CORBA::string_dup(DynamicTC::get_MemberName(L, i+1));
      members[i].type = DynamicTC::get_MemberType(L, i+1);
      members[i].access = DynamicTC::get_MemberAccess(L, i+1);
      members[i].type_def = CORBA::IDLType::_nil();
      lua_pop(L, 1);
   }

   lua_pop(L, 1); // remove "members" table from stack
   return loCfg->orb->create_value_tc(id,name,type_modifier,concrete_base,members);
}

CORBA(TypeCode_ptr) DynamicTC::create_value_box(lua_State *L)
{
   luaL_checktype(L, 2, LUA_TTABLE);
   const char *id = DynamicTC::get_RepID(L, 2);
   const char *name  = DynamicTC::get_Name(L, 2);
   CORBA(TypeCode_ptr) boxed_type = DynamicTC::get_TypeCodeField(L, 2, "boxed_type");
   return loCfg->orb->create_value_box_tc(id,name,boxed_type);
}

CORBA(TypeCode_ptr) DynamicTC::create_interface(lua_State *L, CORBA(TCKind) tk)
{
   luaL_checktype(L, 2, LUA_TTABLE);
   const char *id = DynamicTC::get_RepID(L, 2);
   const char *name  = DynamicTC::get_Name(L, 2);

   if (tk == CORBA(tk_native))
      return loCfg->orb->create_native_tc(id,name);
   else
      return loCfg->orb->create_abstract_interface_tc(id,name);
}

CORBA(TypeCode_ptr) DynamicTC::create_recursive(lua_State *L)
{
   luaL_checktype(L, 2, LUA_TTABLE);
   const char *id = DynamicTC::get_RepID(L, 2);
   return loCfg->orb->create_recursive_tc(id);
}
#endif

#ifndef CORBA_2_3
CORBA(TypeCode_ptr) DynamicTC::create_recursive_sequence(lua_State *L) 
{
   luaL_checktype(L, 2, LUA_TTABLE);
   unsigned long bound = DynamicTC::get_NumericField(L, 2, "bound");
   unsigned long offset = DynamicTC::get_NumericField(L, 2, "offset");
   return loCfg->orb->create_recursive_sequence_tc(bound,offset);
}
#endif

// Auxiliary methods

unsigned long DynamicTC::get_NumericField(lua_State *L, int narg, const char* field)
{
   lua_pushstring(L, field);
   lua_gettable(L, narg);

   if (!lua_isnumber(L, -1))
      luaL_verror(L, "'%s' is missing or non-numeric",field);

   unsigned long numfield = (unsigned long) lua_tonumber(L, -1);
   lua_pop(L, 1);
   return numfield;
}

CORBA(TypeCode_ptr) DynamicTC::get_TypeCodeField(lua_State *L, int narg, 
                                                 const char* field)
{
   lua_pushstring(L, field);
   lua_gettable(L, narg);

   LuaTypeCode* ltc = LuaTypeCode::from_lua(L, lua_gettop(L));
   if (ltc == NULL)
      luaL_verror(L, "'%s' isn't a CORBA::TypeCode Object",field);
   
   lua_pop(L, 1);
   return ltc->get_tc();
}

const char* DynamicTC::get_RepID(lua_State *L, int narg)
{
   lua_pushstring(L, "id");
   lua_gettable(L, narg);

   const char *id = "";
   if (!lua_isnil(L, -1))
   {
     id = lua_tostring(L, -1);
     if ( (id == NULL) || (!LuaOrbUtil::isRepositoryId(id)))
        lua_error(L, "invalid Repository ID");
   }
   lua_pop(L, 1);
   return id;
}

const char* DynamicTC::get_Name(lua_State *L, int narg)
{
   const char *name;
   lua_pushstring(L, "name");
   lua_gettable(L, narg);

   if (lua_isnil(L, -1))
      name = "";
   else
     if (!lua_isstring(L, -1))
        lua_error(L, "invalid TypeCode name");
     else
        name = lua_tostring(L, -1);

   lua_pop(L, 1);
   return name;
}

void DynamicTC::get_Members(lua_State *L, int narg)
{
   lua_pushstring(L, "members");
   lua_gettable(L, narg);

   if (!lua_istable(L, -1)) 
        lua_error(L, "missing or invalid 'members' list");
}

const char* DynamicTC::get_MemberName(lua_State *L, int member)
{
   // member description (table) at the top of the stack
   lua_pushstring(L, "name");
   lua_gettable(L, -2);

   if (!lua_isstring(L, -1))
        luaL_verror(L, "invalid member description (#%d): name", member);

   const char *name = lua_tostring(L, -1);
   lua_pop(L, 1);
   return name;
}

CORBA(TypeCode_ptr) DynamicTC::get_MemberType(lua_State *L, int member)
{
   // member description (table) at the top of the stack
   lua_pushstring(L, "type");
   lua_gettable(L, -2);

   LuaTypeCode* ltc = LuaTypeCode::from_lua(L, lua_gettop(L));
   if (ltc == NULL)
      luaL_verror(L, "invalid member description (#%d): type", member);

   lua_pop(L,1);
   return CORBA::TypeCode::_duplicate(ltc->get_tc());
}

CORBA::Any DynamicTC::get_MemberLabel(lua_State *L, int member,
                                      CORBA(TypeCode_ptr) discriminator_type)
{
   // member description (table) at the top of the stack
   lua_pushstring(L, "label");
   lua_gettable(L, -2);

   if (lua_isnil(L, -1) && discriminator_type->kind() != CORBA(tk_boolean))
        luaL_verror(L, "invalid member description (#%d): label", member);

   CORBA::Any anyValue;
   if (lua_isstring(L, -1))
   {
      const char *label = lua_tostring(L, -1);
      if (strcmp(label, "default") == 0 )
      {
         lua_pop(L, 1);
         CORBA::Any anyValue;
         anyValue <<= CORBA::Any::from_octet((CORBA::Octet) 0);
         return anyValue;
      }
      else
      {
         if (discriminator_type->kind() == CORBA(tk_boolean) &&
             (strcmp(label,"FALSE") == 0 || strcmp(label,"false") == 0))
         {
            lua_pop(L, 1);
            CORBA::Any anyValue;
            anyValue <<= CORBA::Any::from_boolean((CORBA::Boolean) 0);
            return anyValue;
         }
      }
   }
   CORBA::Any* anyPtr = lua2any(L, lua_gettop(L), discriminator_type);
   lua_pop(L, 1);
   return *anyPtr;
}

#ifdef CORBA_2_3
CORBA(ValueModifier) DynamicTC::get_TypeModifier(lua_State *L, int narg)
{
   lua_pushstring(L, "type_modifier");
   lua_gettable(L, narg);

   if (!lua_isstring(L, -1))
      lua_error(L, "invalid Type Modifier");
   const char *vm = lua_tostring(L, -1);
   lua_pop(L, 1);

   for ( unsigned int i = 0; i < sizeof(value_modifiers)/sizeof(char*) ; i++ )
      if (strcmp(vm, value_modifiers[i]) == 0)
         return (CORBA(ValueModifier)) i;

   lua_error(L, "invalid Type Modifier");
   return (CORBA(ValueModifier)) 0;
}

CORBA(Visibility) DynamicTC::get_MemberAccess(lua_State *L, int member)
{
   // member description (table) at the top of the stack
   lua_pushstring(L, "access");
   lua_gettable(L, -2);

   if (!lua_isstring(L, -1))
        luaL_verror(L, "invalid member description (#%d): access", member);

   const char *vis = lua_tostring(L, -1);
   lua_pop(L, 1);

   for ( unsigned int access = 0; access < sizeof(visibility)/sizeof(char*) ; access++ )
      if (strcmp(vis, visibility[access]) == 0)
         return (CORBA(Visibility)) access;

   luaL_verror(L, "invalid member description (#%d): access", member);
   return (CORBA(Visibility)) 0;
}
#endif
