//*****************************************************************************
//
// LuaOrb Library
// Computer Science Department of PUC-Rio
// Rio de Janeiro, Brazil
//
// $Id: luawrap.cpp,v 1.3.2.1 2001/08/07 13:33:03 moura Exp $
//
//*****************************************************************************

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
extern "C"
{
#include <lua.h>
#include <lauxlib.h>
}
#include "luawrap.h"

static void register_methods(lua_State *L, int pos, struct luaL_reg* list, int n)
{
   for (int i = 0; i < n; i++)
   {
      lua_pushstring(L, list[i].name);
      lua_pushcfunction(L, list[i].func);
      lua_settable(L, pos);
   }
}

static void register_tagmethods(lua_State *L, int tag, struct luaL_reg* table, int n)
{
   for (int i=0; i < n; i++)
   {
      lua_pushcfunction(L, table[i].func);
      lua_settagmethod(L, tag, table[i].name);
   }
}

const char* LuaWrapper::reg_WRAPPER_TAG = "LuaWrapper_TAG";

LuaWrapper* LuaWrapper::createWrapper(lua_State *L, const char* name)
{
   int top = lua_gettop(L);

   // check if LuaWrapper_TAG already created in registry
   int WTAG;
   lua_getregistry(L);
   lua_pushstring(L, reg_WRAPPER_TAG);
   lua_gettable(L, -2);
   if (lua_isnil(L, -1))
   {
      WTAG = lua_newtag(L);
      lua_pushstring(L, reg_WRAPPER_TAG);
      lua_pushnumber(L, WTAG);
      lua_settable(L, top+1);		// set in registry table
   }
   else
      WTAG = (int) lua_tonumber(L, -1);

   LuaWrapper* lwrap = new LuaWrapper(L, name);
   
   // Store new wrapper in Lua Registry
   lua_pushstring(L, name);
   lua_pushusertag(L, (void*) lwrap, WTAG);
   lua_settable(L, top+1);

   lua_settop(L, top);
   return lwrap;
}

LuaWrapper* LuaWrapper::getWrapper(lua_State *L, const char* name)
{
   lua_getregistry(L);
   lua_pushstring(L, name);
   lua_gettable(L, -2);
   LuaWrapper* lwrap = (LuaWrapper*) lua_touserdata(L, -1);

   lua_pop(L, 2);	/* pop wrapper and registry */
   return lwrap;
}

void LuaWrapper::deleteWrapper(lua_State *L, const char* name)
{
   lua_getregistry(L);
   lua_pushstring(L, name);
   lua_gettable(L, -2);

   LuaWrapper* lwrap = (LuaWrapper*) lua_touserdata(L, -1);
   if (lwrap)
     delete lwrap;

   lua_pushstring(L, name);
   lua_pushnil(L);
   lua_settable(L, -4);		/* previous value still above registry! */
   lua_pop(L, 2);
}

LuaWrapper::LuaWrapper(lua_State *L, const char* name)
{
   this->mlist = NULL;

   char lua_name[50];
   sprintf(lua_name,"LB_%s_TAG",name);
   this->tag_name = strdup(lua_name);

   sprintf(lua_name,"LB_%s_UDTAG",name);

   this->tag = lua_newtag(L);
   lua_pushnumber(L, tag);
   lua_setglobal(L, tag_name);

   this->ud_tag = lua_newtag(L);
   lua_pushnumber(L, ud_tag);
   lua_setglobal(L, lua_name);
}

LuaWrapper::~LuaWrapper()
{
   free(this->tag_name);
   if (this->mlist)
      delete this->mlist;
}

void LuaWrapper::reg_tagmethods(lua_State *L, struct luaL_reg* table, int n)
{
   register_tagmethods(L, this->tag,table,n);
}

void LuaWrapper::reg_ud_tagmethods(lua_State *L, struct luaL_reg* table, int n)
{
   register_tagmethods(L, this->ud_tag,table,n);
}

void LuaWrapper::reg_methods(struct luaL_reg* table, int n)
{
   if (this->mlist == NULL)
   {
      this->mlist = new LuaWrapper::MethodsList;
   }
   this->mlist->n = n;
   this->mlist->table = table;
}

void LuaWrapper::push(lua_State *L,  void* userdata )
{
   lua_newtable(L);

   lua_pushstring(L, "_USERDATA_REF_");
   lua_pushusertag(L, userdata,this->ud_tag);
   lua_settable(L, -3);

   if (this->mlist)
      register_methods(L, lua_gettop(L), this->mlist->table, this->mlist->n);
   lua_settag(L, this->tag); // Lua table returned at the top of the stack
}

void* LuaWrapper::check_tag(lua_State *L, int narg)
{
   void* userdata = this->from_lua(L, narg);
   luaL_arg_check(L, (userdata!=NULL),narg,"Object type is wrong");
   return userdata;
}

void* LuaWrapper::from_lua(lua_State *L, int pos)
{
   if (lua_istable(L, pos) && lua_tag(L, pos) == this->tag)
   {
      lua_pushstring(L, "_USERDATA_REF_");
      lua_gettable(L, pos);
      void* userdata = lua_touserdata(L, -1);
      lua_pop(L, 1);
      return userdata;
   }
   else
      return NULL;
}
